from .base import FrameField
from ...mesh.mesh_attributes import ArrayAttribute, Attribute, Attribute
from ...mesh.datatypes import *
from ... import operators, utils, attributes, processing
from...utils import maths
from ..features import FeatureEdgeDetector
from ..connection import SurfaceConnectionVertices, FlatConnectionVertices
from ... import geometry as geom
from ...optimize import inverse_power_method

from math import pi
import numpy as np
import cmath
import scipy.sparse as sp
from scipy.sparse import linalg
from osqp import OSQP

class _BaseFrameField2DVertices(FrameField):
    """
    Base class for any frame field defined on the vertices of a surface mesh. Is not meant to be instanciated as is.
    """

    @allowed_mesh_types(SurfaceMesh)
    def __init__(self, 
        supporting_mesh : SurfaceMesh, 
        order : int,
        features : bool = True,
        verbose : bool = True,
        **kwargs
    ):
        """
        Parameters:
            supporting_mesh (SurfaceMesh): the mesh (surface) on which to calculate the frame field
            order (int, optional): Order of the frame field (number of branches). Defaults to 4.
            feature_edges (bool, optional): _description_. Defaults to False.
            verbose (bool, optional): _description_. Defaults to True.
        """
        super().__init__("vertices", verbose=verbose)
        self.mesh : SurfaceMesh = supporting_mesh
        self.order : int = order
        
        self.use_cotan : bool = kwargs.get("use_cotan",True)
        self.cad_correction : bool = kwargs.get("cad_correction", False)
        self.features : bool = features or self.cad_correction # whether feature edges are enabled or not
        self.n_smooth : int = kwargs.get("n_smooth", 10)
        self.smooth_attach_weight = kwargs.get("smooth_attach_weight", None) # either None or the provided value
        self.conn = kwargs.get("custom_connection", None) # (A,B) -> direction (angle) of edge (A,B) in local basis of A
        self.feat : FeatureEdgeDetector = kwargs.get("custom_features", None) # either a FeatureEdgeDetector or None

        self.smooth_normals : bool = kwargs.get("smooth_normals", True) # controls which direction to lock on features and boundary (average of features or first one)

        self.vnormals : ArrayAttribute = None # local basis Z vector (normal)

        self.angles : Attribute = None # angles of every triangle corner
        self.defect : Attribute = None # sum of angles around a vertex (/!\ not 'real' defect which is 2*pi - this)
        self.cot : Attribute = None

        self.var = np.zeros(len(self.mesh.vertices), dtype=complex)

    def _initialize_attributes(self):
        self.angles = attributes.corner_angles(self.mesh)
        self.cot = attributes.cotangent(self.mesh) # re-uses angle attribute
        self.vnormals = attributes.vertex_normals(self.mesh, interpolation="angle") # re-uses angle attribute
        # build defects from angles: 
        self.defect = ArrayAttribute(float, len(self.mesh.vertices))
        for iC, C in enumerate(self.mesh.face_corners):
            self.defect[C] = self.defect[C] + self.angles[iC]
        if self.feat is None:
            self.feat = FeatureEdgeDetector(only_border = not self.features, corner_order=self.order, verbose=self.verbose)
            self.feat.run(self.mesh)
        self.conn = self.conn or SurfaceConnectionVertices(self.mesh, self.feat, vnormal=self.vnormals, angles=self.angles)
    
    def _initialize_variables(self):
        """ Init self.var for feature vertices """
        if self.smooth_normals and self.order%2 != 1:
            for e in self.feat.feature_edges:
                A,B = self.mesh.edges[e]
                edge = self.mesh.vertices[B] - self.mesh.vertices[A]
                vx,vy = self.conn.project(edge,B)
                v = complex(vx, vy)
                vpow = (v/abs(v)) ** self.order
                if abs(self.var[B] + vpow)>1e-10:
                    self.var[B] += vpow

                vx,vy = self.conn.project(edge,A)
                v = complex(vx, vy)
                vpow = (v/abs(v)) ** self.order
                if abs(self.var[A] + vpow)>1e-10:
                    self.var[A] += vpow
        else:
            for e in self.feat.feature_edges:
                A,B = self.mesh.edges[e]
                self.var[A] += cmath.rect(1,self.conn.transport(A,B))**self.order
                self.var[B] += cmath.rect(1,self.conn.transport(B,A))**self.order

        for A in self.feat.feature_vertices:
            if abs(self.var[A])>1e-8:
                self.var[A] /= abs(self.var[A])

    def _compute_attach_weight(self, A, fail_value=1e-3):
        # A is area weight matrix
        lap_no_pt = operators.laplacian(self.mesh)
        try:
            eigs = sp.linalg.eigsh(lap_no_pt, k=2, M=A, which="SM", tol=1e-3, maxiter=1000, return_eigenvectors=False)
        except Exception as e:
            try:
                self.log("First estimation of alpha failed: {}".format(e))
                eigs = sp.linalg.eigsh(lap_no_pt+0.1*sp.identity(lap_no_pt.shape[0]), M=A, k=2, which="SM", tol=1e-3, maxiter=1000, return_eigenvectors=False)
            except:
                self.log("Second estimation of alpha failed: taking alpha = ", fail_value)
                return fail_value
        eigs_non_zero = [e for e in eigs if abs(e)>1e-6]
        if len(eigs_non_zero)==0:
            return fail_value
        return abs(min(eigs_non_zero))


    def flag_singularities(self, singul_attr_name:str = "singuls"):
        """Compute singularity data.

        Creates 3 attributes:
            - an attribute "curvature" on faces storing the gaussian curvature (used as a corrective term for computing the values of the singularities)
            - an attribute "<singul_attr_name>" on faces storing the value (+- 1) of singularities (eventually 0 for non singular triangles)
            - an attribute "angles" on edges storing the angle of the edge, given as the difference between the two frames
        """
        self._check_init()
        curvature = attributes.parallel_transport_curvature(self.mesh, self.conn, persistent=False)
        ZERO_THRESHOLD = 1e-2

        edge_rot = dict() # the rotation induced by the frame field on every edge
        if self.mesh.edges.has_attribute("angles"):
            edge_rot_attr = self.mesh.edges.get_attribute("angles")
            edge_rot_attr.clear()
        else:
            edge_rot_attr = self.mesh.edges.create_attribute("angles", float, 1)
        for ie,(A,B) in enumerate(self.mesh.edges):
            fA,fB = self.var[A], self.var[B] # representation complex for A and B
            aA,aB = self.conn.transport(A,B), self.conn.transport(B,A) # local basis orientation for A and B
            uB = maths.roots(fB, self.order)[0]
            abs_angles = [abs(maths.angle_diff( cmath.phase(uB) - aB - pi, cmath.phase(uA)-aA)) for uA in maths.roots(fA, self.order)]
            angles = [maths.angle_diff( cmath.phase(uB) - aB - pi, cmath.phase(uA)-aA) for uA in maths.roots(fA, self.order)]
            i_angle = np.argmin(abs_angles)
            edge_rot[(A,B)] = angles[i_angle]
            edge_rot[(B,A)] = -angles[i_angle]
            edge_rot_attr[ie] = -angles[i_angle]

        if self.mesh.faces.has_attribute(singul_attr_name):
            singuls = self.mesh.faces.get_attribute(singul_attr_name)
        else:
            singuls = self.mesh.faces.create_attribute(singul_attr_name, int)
        for id_face,(A,B,C) in enumerate(self.mesh.faces):
            angle = 0
            for u,v in [(A,B), (B,C), (C,A)]:
                angle += edge_rot[(u,v)]
            angle += curvature[id_face]
            if angle>ZERO_THRESHOLD:
                singuls[id_face] = 1
            elif angle<-ZERO_THRESHOLD:
                singuls[id_face] = -1

    def export_as_mesh(self, repr_vector=False) -> PolyLine:
        """
        Exports the frame field as a mesh for visualization.

        Returns:
            PolyLine: the frame field as a mesh object
        """
        self._check_init()
        FFMesh = PolyLine()
        L = attributes.mean_edge_length(self.mesh)/3
        n = self.order + 1
        for id_vertex, P in enumerate(self.mesh.vertices):
            E,_ = self.conn.base(id_vertex)
            N = self.vnormals[id_vertex]
            if repr_vector:
                # Representation vector only: 
                angle = cmath.phase(self.var[id_vertex])
                r = geom.rotate_around_axis(E, N, angle)
                p = P + L*r
                FFMesh.vertices += [P, p]
                FFMesh.edges.append((2*id_vertex, 2*id_vertex+1))
            else:
                # Representation of the whole frame
                angle = cmath.phase(self.var[id_vertex])/self.order
                cmplx = [geom.rotate_around_axis(E, N, angle + 2*k*pi/self.order) for k in range(self.order)]
                pts = [P + abs(self.var[id_vertex])*L*r for r in cmplx]
                FFMesh.vertices.append(P)
                FFMesh.vertices += pts
                FFMesh.edges += [(n*id_vertex, n*id_vertex+k) for k in range(1,n)]
        return FFMesh

class FrameField2DVertices(_BaseFrameField2DVertices):
    """
    n-RoSy frame field defined on the vertices of a surface mesh
    """

    @allowed_mesh_types(SurfaceMesh)
    def __init__(self, 
        supporting_mesh : SurfaceMesh, 
        order : int = 4, 
        feature_edges : bool = False, 
        verbose=True,
        **kwargs
    ):
        """
        Parameters:
            supporting_mesh (SurfaceMesh): the mesh (surface) on which to calculate the frame field
            order (int, optional): Order of the frame field (number of branches). Defaults to 4.
            feature_edges (bool, optional): _description_. Defaults to False.
            verbose (bool, optional): _description_. Defaults to True.
        """
        super().__init__(supporting_mesh, order, feature_edges, verbose, **kwargs)
                
    def initialize(self):
        self._initialize_attributes()
        self._initialize_variables()
        if self.cad_correction:
            self._modify_parallel_transport()
        self.initialized = True

    def _modify_parallel_transport(self):
        """
        Implementation of 'Frame Fields for CAD models', Desobry et al, 2021.

        This frame field on vertices has a modified parallel transport so avoid placing singularities in very sharp corners of the mesh.
        """
        curvature = attributes.parallel_transport_curvature(self.mesh, self.conn, persistent=False)
        target_w = dict()

        for e in self.feat.feature_edges:
            A,B = self.mesh.edges[e]
            aA,aB = self.conn.transport(A,B), self.conn.transport(B,A) # local basis orientation for A and B
            fA = self.var[A] # representation complex for frame field at A
            fB = self.var[B] # representation complex for frame field at B
            uB = maths.roots(fB, self.order)[0]
            abs_angles = [abs(maths.angle_diff( cmath.phase(uB) - aB - pi, cmath.phase(uA)-aA)) for uA in maths.roots(fA, self.order)]
            angles = [maths.angle_diff( cmath.phase(uB) - aB - pi, cmath.phase(uA)-aA) for uA in maths.roots(fA, self.order)]
            i_angle = np.argmin(abs_angles)
            target_w[e] = angles[i_angle]

        ## build start w for rotation penalty energy
        cstrfaces = attributes.faces_near_border(self.mesh, 5)
        # 1) constraints
        ncstr = len(self.feat.feature_edges)
        nvar = len(self.mesh.edges)
        Cstr = sp.lil_matrix((ncstr, nvar))
        Rhs = np.zeros(ncstr)
        for i,e in enumerate(self.feat.feature_edges):
            Cstr[i,e] = 1
            Rhs[i] = target_w[e]

        # objective
        D1 = sp.lil_matrix((len(self.mesh.faces), nvar))
        b1 = np.zeros(len(self.mesh.faces))
        for iT,T in enumerate(self.mesh.faces):
            A,B,C = T
            for u,v in [(A,B), (B,C), (C,A)]:
                e = self.mesh.connectivity.edge_id(u,v)
                D1[iT, e]= 1 if u<v else -1
            b1[iT] = -curvature[iT]
        
        D2 = sp.lil_matrix((len(cstrfaces), nvar))
        b2 = np.zeros(len(cstrfaces))
        for i,T in enumerate(cstrfaces):
            A,B,C = self.mesh.faces[T]
            for u,v in [(A,B), (B,C), (C,A)]:
                e = self.mesh.connectivity.edge_id(u,v)
                D2[i,e] = 1e3 if (u<v) else -1e3
            b2[i] = -1e3*curvature[T]

        I = sp.identity(len(self.mesh.edges), format="csc")
        b3 = np.zeros(len(self.mesh.edges))

        P = sp.vstack((D1,D2,I))
        b = np.concatenate([b1,b2,b3])
        b = P.transpose().dot(b)
        P = P.transpose().dot(P)
        osqp_instance = OSQP()
        osqp_instance.setup(P, b, A=Cstr.tocsc(), l=Rhs, u=Rhs, verbose=False)
        res = osqp_instance.solve().x
        for e,(A,B) in enumerate(self.mesh.edges):
            self.conn._transport[(A,B)] +=  res[e]/2
            self.conn._transport[(B,A)] -=  res[e]/2

    def optimize(self):
        self._check_init()
        self.log("Build laplacian operator")
        lap = operators.laplacian(self.mesh, cotan=self.use_cotan, connection=self.conn, order=self.order)
        A = operators.area_weight_matrix(self.mesh).tocsc()

        if len(self.feat.feature_vertices)>0: # We have a border / feature elements -> linear solve
            self.log("Feature element detected (border and/or feature edges)")

            # Build fixed/free vertex partition
            fixedInds,freeInds = [],[]
            for v in self.mesh.id_vertices:
                if v in self.feat.feature_vertices:
                    fixedInds.append(v)
                else:
                    freeInds.append(v)
            
            lapI = lap[freeInds,:][:,freeInds]
            lapB = lap[freeInds,:][:,fixedInds]
            AI = A[freeInds, :][:,freeInds]
            # for lapI and lapB, only lines of freeInds are relevant : lines of fixedInds link fixed variables -> useless constraints
            valB = lapB.dot(self.var[fixedInds]) # right hand side

            self.log("Initial solve of linear system")
            res = linalg.spsolve(lapI, -valB) # first system solved without diffusion
            self.var[freeInds] = res

            if self.n_smooth>0:
                self.log(f"Solve linear system {self.n_smooth} times with diffusion")
                alpha = self.smooth_attach_weight or self._compute_attach_weight(A) # Compute attach weight as smallest eigenvalue of the laplacian if not provided as argument
                self.log("Attach weight: {}".format(alpha))
                mat = lapI - alpha * AI.astype(complex)
                for _ in range(self.n_smooth):
                    self.normalize()
                    valI2 = alpha * AI.dot(self.var[freeInds])
                    res = linalg.spsolve(mat, - valB - valI2)
                    self.var[freeInds] = res
            self.normalize()

        else: # No border -> eigensolve
            self.log("No border detected")
            self.log("Initial solve of linear system using an eigensolver")
            self.var = inverse_power_method(lap,A)
            if self.n_smooth>0:
                self.log(f"Solve linear system {self.n_smooth} times with diffusion")
                alpha = self.smooth_attach_weight or self._compute_attach_weight(A) # Compute attach weight as smallest eigenvalue of the laplacian
                self.log("Attach weight: {}".format(alpha))
                mat = lap  - alpha * A.astype(complex)
                for _ in range(self.n_smooth):
                    self.normalize()
                    valI2 = alpha * A.dot(self.var)
                    self.var = linalg.spsolve(mat, - valI2)
            self.normalize()
        self.smoothed = True

class TrivialConnectionVertices(_BaseFrameField2DVertices):
    """
    Implementation of 'Trivial Connections on Discrete Surfaces' by Keenan Crane and Mathieu Desbrun and Peter Schröder, 2010
    
    A frame field on vertices that computes the smoothest possible frame field with prescribed singularity cones at some vertices.
    Does not constraint non-contractible cycles
    """

    @allowed_mesh_types(SurfaceMesh)
    def __init__(self, 
        supporting_mesh : SurfaceMesh, 
        singus_indices:Attribute, 
        order:int = 4, 
        verbose:bool=True,
        **kwargs):
        super().__init__(supporting_mesh, order, feature_edges=False, verbose=verbose, **kwargs)
        self.singus = singus_indices
        self.rotations : np.ndarray = None

    def initialize(self):
        self._initialize_attributes()
        self.var = np.zeros(len(self.mesh.vertices), dtype=complex)
        self.initialized = True
        
    def optimize(self):
        nvar = len(self.mesh.edges)
        ncstr = len(self.mesh.faces)
        curvature = attributes.parallel_transport_curvature(self.mesh, self.conn, persistent=False)
        CstM = sp.lil_matrix((ncstr,nvar))
        CstX = np.zeros(ncstr)
        for F,face in enumerate(self.mesh.faces):
            for u,v in utils.cyclic_pairs(face):
                e = self.mesh.connectivity.edge_id(u,v)
                CstM[F,e] = 1 if (u<v) else -1
            CstX[F] = curvature[F] - 2* pi * self.singus[F] / self.order
        CstM = CstM.tocsc()
        A = sp.eye(nvar, format="csc")
        instance = OSQP()
        instance.setup(P=A, q=None, A=CstM, u=CstX, l=CstX)
        res = instance.solve()
        self.rotations = res.x

        tree = processing.trees.EdgeSpanningTree(self.mesh)()
        for vertex,parent in tree.traverse():
            if parent is None:
                self.var[vertex] = complex(1.,0.)
                continue
            zf = self.var[parent]
            e = self.mesh.connectivity.edge_id(parent,vertex)
            pt = self.conn.transport(vertex,parent) - self.conn.transport(parent,vertex) + pi
            w = self.rotations[e] if vertex<parent else -self.rotations[e]
            zv = zf * cmath.rect(1, 4*(w + pt))
            self.var[vertex] = zv
        self.smoothed = True


import numpy as np
from scipy.spatial import KDTree
from enum import Enum

from ..mesh.datatypes import *
from ..mesh.mesh_data import RawMeshData
from ..mesh.mesh_attributes import ArrayAttribute
from .worker import Worker
from .. import geometry as geom
from ..geometry import Vec
from .. import utils

class PointCloudNormalEstimator(Worker):
    """An estimator of normal directions for unstructured point clouds, based on k-nearest neighbors and singular value decomposition of their correlation matrix.


    References:
        [1] Surface reconstruction from unorganized points, Hoppe et al., 1992

    Example:
        [https://github.com/GCoiffier/mouette/blob/main/examples/point_cloud_normal_estimation.py](https://github.com/GCoiffier/mouette/blob/main/examples/point_cloud_normal_estimation.py)
    """

    class OrientStrategy(Enum):
        NONE = 0
        SPANNING_TREE = 1

        @classmethod
        def from_string(cls, txt : str):
            txt = txt.lower()
            if "mst" in txt : return cls.SPANNING_TREE
            return cls.NONE
        

    def __init__(self, 
        n_neighbors : int = 5,
        save_on_pc : bool = True,
        compute_curvature : bool = False,
        orientation_mode : str = "mst",
        verbose = True, 
        **kwargs):
        """
        Args:
            n_neighbors (int, optional): number of neighbor points to consider for local plane fitting. Defaults to 5.
            save_on_pc (bool, optional): whether to store the resulting normals onto the point cloud as a vertex attribute, or as independent arrays. Defaults to True.
            compute_curvature (bool, optional): whether to also compute an estimate of curvature on the point cloud. This curvature estimate is computed as the ratio of the smallest eigenvalue and the sum of eigenvalues of the correlation matrix at each point (between 0 and 1). Defaults to False.
            orientation_mode (str, optional): Heuristic strategy for consistent orientation of normals. Choices are ["None", "mst"] Defaults to "mst".
            verbose (bool, optional): verbose mode. Defaults to True.


        Keyword Args:
            normal_attribute_name (str): The name of the attribute in which normals are stored. Ignored if save_on_pc is set to False. Defaults to "normals".
            curvature_attribute_name (str): The name of the attribute in which the curvature estimation is stored. Ignored if save_on_pc is set to False. Defaults to "curvature".
        """
        super().__init__("PCNormalEstimation", verbose)
        self._n_neighbors : int = n_neighbors
        self._save_on_pc : bool = save_on_pc
        self._orientation_mode = PointCloudNormalEstimator.OrientStrategy.from_string(orientation_mode)
        
        self._normal_attr_name = kwargs.get("normal_attribute_name", "normals")
        self._curvature_attr_name = kwargs.get("curvature_attribute_name", "curvature")
        self._compute_curvature : bool = compute_curvature
        
        self.pc : PointCloud = None
        self.normals : ArrayAttribute = None
        self.curvature : ArrayAttribute = None

    def clear(self):
        self.pc = None
        self.normals = None
        self.curvature = None

    def normals_as_vector_field(self, scale:float = 1.):
        if self.normals is None : return None 
        pl = RawMeshData()
        for i in self.pc.id_vertices:
            P1 = self.pc.vertices[i]
            P2 = self.pc.vertices[i] + scale * self.normals[i]
            pl.vertices += [P1, P2]
            pl.edges.append((2*i,2*i+1))
        return PolyLine(pl)

    @allowed_mesh_types(PointCloud)
    def run(self, point_cloud : PointCloud, **kwargs):

        ### Prepare data and containers
        self.log("Prepare data and containers")
        self.pc = point_cloud
        points = np.asarray(point_cloud.vertices)
        
        kdtree = kwargs.get("kdtree", None)
        if kdtree is not None:
            utils.check_argument("kdtree", kdtree, KDTree)
        else:
            self.log("Compute kd-tree")
            kdtree = KDTree(points)
        _,KNN = kdtree.query(points, self._n_neighbors+1)

        if self._save_on_pc:
            self.normals = point_cloud.vertices.create_attribute(self._normal_attr_name, float, 3, dense=True)
            self.curvature = point_cloud.vertices.create_attribute(self._curvature_attr_name, float, 1, dense=True)
        else:
            self.normals = np.zeros((len(point_cloud.vertices), 3))
            self.curvature = np.zeros(len(point_cloud.vertices))

        ### Run normal estimation
        self.log("Estimate normals with SVD")
        self._estimate_normals_svd(KNN)

        ### Consistently orient the normals
        if self._orientation_mode == PointCloudNormalEstimator.OrientStrategy.SPANNING_TREE:
            self.log("Consistently orient normals (MST strategy)")
            self._correct_normal_orientation_mst(kdtree, KNN)

    def _estimate_normals_svd(self, KNN):
        points = np.asarray(self.pc.vertices)
        for i in self.pc.id_vertices:
            pts_i = points[KNN[i,:],:]
            center = np.mean(pts_i, axis=0)
            mat_i = pts_i-center
            svd = np.linalg.svd(mat_i.T, full_matrices=False)
            self.normals[i] = svd[0][:,-1]
            if self._compute_curvature :
                l0,l1,l2 = svd[1]
                assert l0 >= l1 >= l2
                self.curvature[i] = l2/(l0+l1+l2)

    def _correct_normal_orientation_mst(self, kdtree, KNN):
        ### Build Minimal Spanning tree
        edges = set()
        for i in self.pc.id_vertices:
            for j in KNN[i,:]:
                if j==i: continue
                edges.add(utils.keyify(i,j))
        pr_queue = utils.PriorityQueue()
        for (i,j) in edges:
            pr_queue.push((i,j), 1 - abs(np.dot(self.normals[i], self.normals[j])))

        UF = utils.UnionFind(self.pc.id_vertices)
        adjacency = [[] for _ in self.pc.id_vertices]
        edges_mst = []
        while not pr_queue.empty():
            elem = pr_queue.pop()
            i,j = elem.x
            if UF.find(i) != UF.find(j):
                UF.union(i,j)
                adjacency[i].append(j)
                adjacency[j].append(i)
                edges_mst.append([i,j])

        ### Set starting orientation
        aabb = geom.AABB.of_mesh(self.pc)
        exterior_pt = aabb.center + Vec(0, 2*aabb.span[1], 0)
        _,top_pt = kdtree.query(exterior_pt)
        if np.dot(self.normals[top_pt], exterior_pt - self.pc.vertices[top_pt])<0.:
            self.normals[top_pt] *= -1 # if this normal is not outward, we are doomed
        
        ### Traverse spanning tree
        to_visit = []
        visited = np.zeros(len(self.pc.vertices), dtype=bool)
        for adj_pt in adjacency[top_pt]:
            to_visit.append((top_pt, adj_pt))
        visited[top_pt] = True
        
        while len(to_visit)>0:
            parent, node = to_visit.pop()
            if visited[node] : continue
            visited[node] = True
            Nv = self.normals[node]
            Np = self.normals[parent] # suppose that this one is correctly oriented
            if np.dot(-Nv,Np)>np.dot(Nv,Np):
                self.normals[node] *= -1
            for child in adjacency[node]:
                if not visited[child]:
                    to_visit.append((node,child))


import numpy as np

def estimate_gradient_finite_differentes(fun, X : np.ndarray, eps: float = 1e-8):
    grad_FD = np.zeros_like(X)
    nvar = X.shape[0]
    for i in range(nvar):
        X[i] += eps
        ep = fun(X)
        X[i] -= 2*eps
        em = fun(X)
        X[i] += eps
        grad_FD[i] = (ep - em)/(2*eps)
    return grad_FD


from ..mesh.datatypes import *
from ..mesh.mesh import merge
from .shapes import cylinder, icosphere
from ..attributes import mean_edge_length

def spherify_vertices(points : PointCloud, radius : float = 1e-2, n_subdiv=1) -> SurfaceMesh:
    """Transforms vertices of a point cloud as icospheres

    Args:
        points (PointCloud): the input point cloud
        radius (float, optional): radius of each sphere. Defaults to 1e-2.
        n_subdiv (int, optional): number of subdivisions of the icospheres. Defaults to 1.

    Returns:
        SurfaceMesh
    """
    spheres = []
    if isinstance(points, Mesh):
        what = points.vertices
    else:
        what = points
    for P in what:
        mp = icosphere(n_subdiv, P, radius)
        spheres.append(mp)
    return merge(spheres)


@forbidden_mesh_types(PointCloud)
def cylindrify_edges( mesh : PolyLine, radius: float = 5e-2, N=50) -> SurfaceMesh:
    """ Transforms edges of a polyline as cylinder surface

    Parameters:
        mesh (PolyLine): the input mesh
        radius (float, optional): Radius of the output cylinders. Defaults to 5e-2.
        N (int, optional): Number of points inside each circle of the cylinders. Defaults to 50.

    Returns:
        SurfaceMesh
    """
    if len(mesh.edges)==0: 
        return SurfaceMesh()

    L = mean_edge_length(mesh)
    cylinders = []
    for A,B in mesh.edges:
        pA,pB = mesh.vertices[A], mesh.vertices[B]
        cylinders.append(cylinder(pA,pB, L*radius, N, fill_caps=False))
    return merge(cylinders)


import cmath
from math import pi, sqrt

def roots(c : complex, pow: int, normalize=True):
    """Given a complex number c, compute and returns the four ^(1/4) roots 
    Parameters:
        c (complex): input complex number
        pow (int): the power of the root. Returns power-th roots of c
        normalize (bool, optional): If True, roots will have module 1. Defaults to True

    Returns:
        list of complex roots
    """
    r,t = cmath.polar(c)
    r = 1 if normalize else r**(1/pow)
    return [cmath.rect(r, (t + 2*k*pi)/pow) for k in range(pow)]

def angle_diff(a : float, b : float) -> float:
    return (a - b + pi) % (2*pi) - pi

def principal_angle(a : float) -> float :
    """
    From an arbitrary angle value, returns the equivalent angle which values lays in [-pi, pi[
    """
    b = a%(2*pi)
    if b>pi:
        b-=2*pi
    return b

def solve_quadratic(A : float, B : float, C : float):
    """
    Solves Ax² + Bx + C = 0 for real-valued roots

    Args:
        A (float): coefficient of X² 
        B (float): coefficient of X
        C (float): constant coefficient

    Returns:
        list: a list containing 0, 1 or 2 roots
    """
    if A==0: # linear case
        if B == 0:
            return []
        return [-C/B]
    delta = B*B-4*A*C
    if delta<0:
        return []
    if abs(delta)<1e-14: # delta = 0
        return [-B/(2*A)]
    else:
        delta = sqrt(delta)
        return [ (-B + delta)/(2*A), (-B - delta)/(2*A)]



from ..mesh_data import RawMeshData
from ..mesh import PointCloud
import numpy as np

def import_xyz(filepath : str) -> RawMeshData:
    """Imports a point cloud from a .xyz file on the disk

    Parameters:
        filepath (str): path to the .xyz file

    Returns:
        RawMeshData: parsed file
    """
    obj = RawMeshData()
    nrmls = []
    with open(filepath, 'r') as f:
        for v in f.readlines():
            data = [float(x) for x in v.strip().split()]
            if len(data)==1: continue
            obj.vertices.append(data[:3])
            if len(data)>=6:
                nrmls.append(data[3:6])
    if len(nrmls)==len(obj.vertices):
        normals = obj.vertices.create_attribute("normals", float, 3, dense=True)
        normals._data = np.array(nrmls)
    return obj

def export_xyz(mesh : PointCloud, filepath : str) -> None:
    """Exports a Point Cloud to a .xyz file

    Parameters:
        mesh (PointCloud): the object to be exported
        filepath (str): path to the .xyz file
    """
    with open(filepath, 'w') as f:
        if mesh.vertices.has_attribute("normals"):
            normals = mesh.vertices.get_attribute("normals")
            for iv,v in enumerate(mesh.vertices):
                nv = normals[iv]
                f.write("{} {} {} {} {} {}\n".format(v[0], v[1], v[2], nv[0], nv[1], nv[2]))
        else:
            for v in mesh.vertices:
                f.write("{} {} {}\n".format(v[0], v[1], v[2]))


from .base import BaseParametrization
from ...mesh.datatypes import *
from ...geometry import Vec
from ... import operators
from ..border import extract_border_cycle
from ...attributes.glob import euler_characteristic
from ...utils.argument_check import check_argument

import numpy as np
from enum import Enum
import cmath
from math import pi
from scipy.sparse import linalg


class TutteEmbedding(BaseParametrization):
    """
    Tutte's embedding parametrization method for a disk inside a fixed boundary.
    The parametrization is locally injective (Floater, 1997) provided the boundary is convex.

    References:
        - [1] _How to draw a graph_, Tutte W.T., 1963
        
        - [2] _Parametrization and smooth approximation of surface triangulations_, Floater M.S., 1997

    Example:
        [https://github.com/GCoiffier/mouette/blob/main/examples/parametrization/tutte.py](https://github.com/GCoiffier/mouette/blob/main/examples/parametrization/tutte.py)
    """

    class BoundaryMode(Enum):
        """
        Enum that represents the shape of the boundary curve.

        Attributes:
            CIRCLE (int): 0
            SQUARE (int): 1
            CUSTOM (int): 2
        """
        
        CIRCLE = 0
        SQUARE = 1
        CUSTOM = 2

        @staticmethod
        def from_string(s :str):
            """
            Initializes the value of the enum from the corresponding string
            
            Raises:
                Exception: fails if the string does not correspond to a valid mode
            """
            if "circle" in s.lower():
                return TutteEmbedding.BoundaryMode.CIRCLE
            if "square" in s.lower():
                return TutteEmbedding.BoundaryMode.SQUARE
            if "custom" in s.lower():
                return TutteEmbedding.BoundaryMode.CUSTOM
            raise Exception(f"'{s}' does correspond to any boundary mode. Choices are : 'circle', 'square', 'custom'")

    def __init__(self, mesh:SurfaceMesh, boundary_mode:str = "circle", use_cotan:bool=False, verbose:bool=False, **kwargs):
        """
        Args:
            mesh (SurfaceMesh): the mesh to embed. Should be a surface with disk topology.
            boundary_mode (str, optional): Shape of the boundary. Possible choices are ["square", "circle"]. Defaults to "circle".
            use_cotan (bool, optional): If True, uses cotangents as weights in the Laplacian matrix [2]. Otherwise, use Tutte's original barycentric embedding [1]. Defaults to False.
            verbose (bool, optional): verbose mode. Defaults to False.
            
        Keyword Args:
            save_on_corners (bool, optional): if True, the resulting uv-coordinates will be stored in an attribute on face corners. Otherwise, they are stored in an attribute on vertices. Defaults to True
            custom_boundary (np.ndarray, optionnal): a Nx2 array containing custom coordinates for the boundary vertices (N being the number of boundary vertices). 
            If provided, the boundary_mode argument is ignored. Defaults to None.

        Attributes:
            uvs (Attribute): an attribute containing the uv-coordinates of the parametrization
            save_on_corners (bool): whether the coordinates are saved as a vertex attribute or a face_corner attribute

        Raises:
            InvalidArgumentValueError : if 'boundary_mode' is not "square" or "circle".
        """
        check_argument("boundary_mode", boundary_mode, str, ["square", "circle"])
        super().__init__("Tutte", mesh, verbose, **kwargs)
        self._custom_bnd : np.ndarray = kwargs.get("custom_boundary", None)
        self._use_cotan : bool = use_cotan
        if self._custom_bnd is None:
            self._bnd_mode = TutteEmbedding.BoundaryMode.from_string(boundary_mode)
        else:
            self._bnd_mode = TutteEmbedding.BoundaryMode.CUSTOM

    def run(self) :
        """Computes the parametrization. Result is stored in the `uvs` attribute

        Raises:
            Exception: fails if the mesh is not a topological disk
        """
        if euler_characteristic(self.mesh)!=1:
            raise Exception("Mesh is not a topological disk. Cannot run parametrization.")
        
        Ubnd,Vbnd = self._initialize_boundary(self._bnd_mode)

        lap = operators.laplacian(self.mesh, cotan=self._use_cotan)
        freeInds = self.mesh.interior_vertices
        if self._bnd_mode == TutteEmbedding.BoundaryMode.CUSTOM:
            bndInds = self.mesh.boundary_vertices
        else: # otherwise boundary should be sorted
            bndInds, _ = extract_border_cycle(self.mesh)

        LI = lap[freeInds, :][:, freeInds]
        LB = lap[freeInds, :][:, bndInds]

        U = linalg.spsolve(LI, -LB.dot(Ubnd))
        V = linalg.spsolve(LI, -LB.dot(Vbnd))

        if self.save_on_corners:
            self.uvs = self.mesh.face_corners.create_attribute("uv_coords", float, 2, dense=True)
            for i,v in enumerate(freeInds):
                for c in self.mesh.connectivity.vertex_to_corners(v):
                    self.uvs[c] = Vec(U[i], V[i])
            for i,v in enumerate(bndInds):
                for c in self.mesh.connectivity.vertex_to_corners(v):
                    self.uvs[c] = Vec(Ubnd[i], Vbnd[i])
        else:
            self.uvs = self.mesh.vertices.create_attribute("uv_coords", float, 2, dense=True)
            for i,v in enumerate(freeInds):
                self.uvs[v] = Vec(U[i], V[i])
            for i,v in enumerate(bndInds):
                self.uvs[v] = Vec(Ubnd[i], Vbnd[i])

    def _initialize_boundary(self, boundary_mode):
        n = len(self.mesh.boundary_vertices)
        U, V = np.zeros(n), np.zeros(n)
        if boundary_mode == TutteEmbedding.BoundaryMode.CUSTOM:
            return self._custom_bnd[:,0], self._custom_bnd[:,1]     
        elif boundary_mode == TutteEmbedding.BoundaryMode.CIRCLE:
            for i in range(n):
                rt = cmath.rect(1., 2*pi*i/n)
                U[i] = rt.real
                V[i] = rt.imag
        elif boundary_mode == TutteEmbedding.BoundaryMode.SQUARE:
            corners = [0, n//4, n//2, (3*n)//4]
            U[corners[0]], V[corners[0]] = 0,0
            U[corners[1]], V[corners[1]] = 1,0
            U[corners[2]], V[corners[2]] = 1,1
            U[corners[3]], V[corners[3]] = 0,1
            for i in range(1, n//4):
                U[i] = 4*i/n
                # V = 0
            for i,v in enumerate(range(n//4+1, n//2)):
                U[v] = 1
                V[v] = 4*i/n
            for i,v in enumerate(range(n//2+1, (3*n)//4)):
                U[v] = 1-4*i/n
                V[v] = 1
            for i,v in enumerate(range((3*n)//4+1, n)):
                # U = 0
                V[v] = 1-4*i/n
        return U, V


from ..mesh.mesh_attributes import Attribute, ArrayAttribute
from ..mesh.datatypes import *
from .misc_corners import cotangent
from .. import geometry as geom
from ..geometry import Vec
import numpy as np

@forbidden_mesh_types(PointCloud)
def edge_length(mesh : Mesh, name = "length", persistent:bool=True, dense:bool = True) -> Attribute:
    """Compute edge lengths across the mesh.

    Parameters:
        mesh (Mesh): the input mesh. PointClouds are forbidden (no edges).
        name (str, optional): name of the attribute. Defaults to "length".
        persistent (bool, optional): If the attribute is persistent (stored in the mesh object) or not. Defaults to True.
        dense (bool, optional): Is the attribute dense (numpy array) or not (dict). Defaults to True
    """
    if persistent:
        length = mesh.edges.create_attribute(name, float, dense=True)
    else:
        length = ArrayAttribute(float, len(mesh.edges)) if dense else Attribute(float)
    for e,(a,b) in enumerate(mesh.edges):
        pA,pB = mesh.vertices[a], mesh.vertices[b]
        length[e] = geom.distance(pA,pB)
    return length

@allowed_mesh_types(SurfaceMesh)
def curvature_matrices(mesh : SurfaceMesh) -> Attribute:
    """
    Curvature matrix for each edge on the mesh, as defined in 'Restricted Delaunay Triangulations and Normal Cycle', David Cohen-Steiner and Jean-Marie Morvan, 2003

    Note:
        See Curvature Frame Fields for their aggregation on triangles or vertices
    """

    data = np.zeros((len(mesh.edges), 3,3), dtype=np.float64)
    for e, (A,B) in enumerate(mesh.edges):
        T1,T2 = mesh.connectivity.edge_to_faces(A,B)
        if T1 is not None and T2 is not None:
            _,_,n1 = geom.face_basis(*(mesh.vertices[u] for u in mesh.faces[T1]))
            _,_,n2 = geom.face_basis(*(mesh.vertices[u] for u in mesh.faces[T2]))
            # angle = geom.signed_angle_2vec3D(n1,n2,n2-n1)
            angle = geom.angle_2vec3D(n1,n2)
            edge = Vec.normalized(mesh.vertices[B] - mesh.vertices[A])
            data[e,:,:] = angle * np.outer(edge,edge)
    return data

@allowed_mesh_types(SurfaceMesh)
def cotan_weights(mesh : SurfaceMesh, name="cotan_weight", persistent:bool = True, dense:bool = True)-> Attribute:
    """ Compute the cotan weights of edges.
    The weight of an edge separating T1 and T2 is the sum of cotangent of opposite angles in T1 and T2

    Parameters:
        mesh (Mesh): the input mesh.
        name (str, optional): name of the attribute. Defaults to "cotan_weight".
        persistent (bool, optional): If the attribute is persistent (stored in the mesh object) or not. Defaults to True.
        dense (bool, optional): Is the attribute dense (numpy array) or not (dict). Defaults to True
    """
    if not mesh.is_triangular():
        raise Exception("Tried to compute cotan weights on a non-triangulated mesh")

    if persistent:
        cw = mesh.edges.create_attribute(name, float, dense=dense)
    else:
        cw = ArrayAttribute(float, len(mesh.edges)) if dense else Attribute(float)

    if mesh.face_corners.has_attribute("cotan"):
        cot = mesh.face_corners.get_attribute("cotan")
    else:
        cot = cotangent(mesh,persistent=persistent)

    for e,(A,B) in enumerate(mesh.edges):
        T,iA,iB = mesh.connectivity.direct_face(A,B,True)
        if T is not None:
            cnr = mesh.connectivity.face_to_first_corner(T)+3-iA-iB
            cw[e] += cot[cnr]/2
        T,iB,iA = mesh.connectivity.direct_face(B,A,True)
        if T is not None:
            cnr = mesh.connectivity.face_to_first_corner(T)+3-iA-iB
            cw[e] += cot[cnr]/2
    return cw 


from .faces2d import _BaseFrameField2DFaces
from ...mesh.datatypes import *
from ...attributes import curvature_matrices
from ... import operators
import numpy as np
from scipy.sparse import linalg

class PrincipalDirectionsFaces(_BaseFrameField2DFaces):
    """
    Principal curvature direction estimation using a frame field on faces.

    Implementation based on 'Restricted Delaunay Triangulations and Normal Cycle',  David Cohen-Steiner and Jean-Marie Morvan, 2003
    """

    def __init__(self, 
        supporting_mesh : SurfaceMesh,
        features : bool = True,
        verbose :bool = False,
        **kwargs):
        """
        Principal curvature direction estimation using a frame field on faces.

        Args:
            supporting_mesh (SurfaceMesh): the surface mesh on which to perform the estimation
            verbose (bool, optional): verbose mode. Defaults to False.
        
        Additionnal arguments:
            TODO
        

        Note:
            Order of the frame field is fixed at 4 since principal curvature directions form an orthonormal basis.
        """
        super().__init__(supporting_mesh, 4, features, verbose, **kwargs)
        self.curv_mat_faces : np.ndarray = None # shape (|T|,3,3)

    def _initialize_curvature(self):
        curv_mat_edges = curvature_matrices(self.mesh)
        adj_e = dict([(e,set()) for e in self.mesh.id_edges])
        for T in self.mesh.id_faces:
            for T2 in self.mesh.connectivity.face_to_faces(T):
                for e in self.mesh.connectivity.face_to_edges(T2):
                    adj_e[e].add(T)

        self.curv_mat_faces = np.zeros((len(self.mesh.faces),3,3))
        for e in self.mesh.id_edges:
            for T in adj_e[e]:
                self.curv_mat_faces[T] += curv_mat_edges[e]

    def initialize(self):
        self.log("Initialize attributes")
        self._initialize_attributes() # /!\ may change mesh combinatorics near boundary
        self.log("Compute curvature matrices")
        self._initialize_curvature()
        self.log("Initialize variables")
        self._initialize_variables()
        self.initialized = True

    def optimize(self):
        for T in self.mesh.id_faces:
            if abs(self.var[T])!=0: continue # already initialized => T is on the boundary
            u,s,_ = np.linalg.svd(self.curv_mat_faces[T,:,:], hermitian=True)
            X,Y = self.conn.base(T)
            # three vectors -> normal and two principal components
            # eigenvalue of normal is 0
            # PC are orthogonal (eigenvects of symmetric matrix) -> we rely on the eigenvect of greatest eigenvalue and take orthog direction
            if s[0]<1e-10 : # zero matrix -> no curvature information -> we take the first basis vector and hope to smooth correctly
                self.var[T] = 0 + 0j
            else:
                # representation vector
                v = u[:,0]
                c = complex(X.dot(v), Y.dot(v))
                self.var[T] = (c/abs(c))**4

        if self.n_smooth > 0:
            # diffuse the curvature results to get a smoother results (especially where curvature was not defined)
            lap = operators.laplacian_triangles(self.mesh, order=4, connection=self.conn).tocsc()
            A = operators.area_weight_matrix_faces(self.mesh)
            self.log(f"Solve linear system {self.n_smooth} times with diffusion")
            alpha = self.smooth_attach_weight or self._compute_attach_weight(A, 0.1)
            A = A.astype(complex)
            self.log("Attach weight: {}".format(alpha))
            # TODO : take boundary into account
            # if len(self.feat.feature_vertices)>0: # we have a boundary
            #     raise NotImplementedError
            mat = lap - alpha * A
            solve = linalg.factorized(mat)
            for _ in range(self.n_smooth):
                valI2 = alpha * A.dot(self.var)
                self.var = solve(-valI2)
                self.normalize()
        self.smoothed = True


from ..mesh.datatypes import *
from ..mesh.mesh_data import DataContainer, RawMeshData
from ..geometry import Vec, AABB
import numpy as np

from dataclasses import dataclass
from enum import Enum
from abc import abstractmethod

# list_to_number = lambda L : reduce(lambda a,b : (a<<1) + int(b), L)

# class _SpacePartitionTree:
    
#     @dataclass
#     class Node:
#         id : int # id of the node
#         depth : int # depth of the node inside the tree
#         parent : int  # id of parent node (-1 if root node)
#         corners : list # indices of vertex corners of the node (should be 4 elements)
#         children : list = None # id of children nodes
#         points: set = None # id of contained points

#     def __init__(self, points : np.ndarray, dim :int, **kwargs):
#         self.dim = dim
#         self.max_depth : int = kwargs.get("max_depth", 8)
#         self.points = DataContainer(id="vertices")
#         self.domain = kwargs.get("domain", None)
#         if self.domain is None or not isinstance(self.domain, AABB) or self.domain.dim != dim:
#             diameter = np.amax(self.points) - np.amin(self.points)
#             self.domain = AABB.of_points(self.points, 0.01 * diameter)

#         self._where = self.points.create_attribute("id", int)

#         self._id_node = -1
#         self._nodes = []

#         ## Build a root node
#         self._create_node(0, -1, [0,1,2,3])

#     def _get_new_node_id(self):
#         self._id_node += 1
#         return self._id_node

#     def _create_node(self, depth, parent, verts):
#         assert len(verts)==4
#         new_id = self._get_new_node_id()
#         new_node = QuadTree.Node(new_id, depth, parent, verts)
#         new_node.points = []
#         new_node.children = []
#         vmin,vmax = self.vertex(verts[0]), self.vertex(verts[2])
#         self._bb.append(AABB((vmin.x, vmin.y), (vmax.x, vmax.y)))
#         self._nodes.append(new_node)
#         return new_id

#     def vertex(self,i):
#         return self._tree_vert.vertices[i]

#     def _subdivide(self, node_id):
#         """
#         Subdivision convention
#         v3 -- v6 -- v2
#         |     |      |
#         |     |      |
#         v7 -- v8 --  v5
#         |     |      |
#         |     |      |
#         v0 -- v4 -- v1 
#         """
#         node : QuadTree.Node = self._nodes[node_id]
#         if len(node.children)!=0 : return # not a leaf -> cannot subdivide
#         nv = len(self._tree_vert.vertices)
#         v0,v1,v2,v3 = node.corners
#         p0,p1,p2,p3 = (self.vertex(v) for v in node.corners)
#         v4,p4 = nv,   (p0+p1)/2
#         v5,p5 = nv+1, (p1+p2)/2
#         v6,p6 = nv+2, (p2+p3)/2
#         v7,p7 = nv+3, (p0+p3)/2
#         v8,p8 = nv+4, (p0+p1+p2+p3)/4
#         self._tree_vert.vertices += [p4,p5,p6,p7,p8]
#         n1,n2,n3,n4 = (
#             self._create_node(node.depth+1, node.id, [v0,v4,v8,v7]),
#             self._create_node(node.depth+1, node.id, [v4,v1,v5,v8]),
#             self._create_node(node.depth+1, node.id, [v8,v5,v2,v6]),
#             self._create_node(node.depth+1, node.id, [v7,v8,v6,v3])
#         )
#         node.children = [n1,n2,n3,n4]
#         for p in node.points:
#             self._insert_point(self.points[p], p, node.id)
#         node.points = []
    
#     def insert_point(self, x:float, y:float):
#         """Adds a point inside the Quadtree Structure
#         """
#         pt = Vec(x, y, 0.)
#         if not self._bb[0].contains_point(pt): return
#         self.points.append(pt)
#         self._insert_point(pt, len(self.points)-1, 0)

#     def _insert_point(self, pt, pt_id, node_id):
#         node : QuadTree.Node = self._nodes[node_id]
#         if len(node.children)!=0:
#             # not a leaf -> recursive call to children
#             for child in node.children:
#                 self._insert_point(pt, pt_id, child)
#         else:
#             BB : AABB = self._bb[node_id]
#             if BB.contains_point(pt):
#                 self._where[pt_id] = node_id
#                 node.points.append(pt_id)
#                 if len(node.points)>QuadTree.MAX_PT_PER_NODE:
#                     # trigger subdivision
#                     self._subdivide(node_id)
    
#     def query(self, pt:Vec):
#         return
    
#     def find_all_points_in(self, box:AABB) -> list:
#         return
    
#     @abstractmethod
#     def export_visualization(self):
#         return
    

# class QuadTree(_SpacePartitionTree):
#     def __init__(self, points : np.ndarray, **kwargs):
#         points = np.array(points)[:,:2]
#         super().__init__(points, 2, **kwargs)

#     def export_visualization(self) -> PolyLine:
#         out = RawMeshData()
#         out.vertices = self._tree_vert.vertices
#         for node in self._nodes:
#             v0,v1,v2,v3 = node.corners
#             out.edges += [(v0,v1), (v1,v2), (v2,v3), (v3,v0)]
#         return PolyLine(out)


# class OcTree(_SpacePartitionTree):
#     def __init__(self, points : np.ndarray, **kwargs):
#         points = np.array(points)[:,:3]
#         super().__init__(points, 3, **kwargs)

#     def export_visualization(self) -> PolyLine:
#         out = RawMeshData()
#         out.vertices = self._tree_vert.vertices
#         for node in self._nodes:
#             v0,v1,v2,v3 = node.corners
#             out.edges += [(v0,v1), (v1,v2), (v2,v3), (v3,v0)]
#         return PolyLine(out)



from ...mesh.datatypes import *
from ...mesh.mesh_attributes import Attribute
from ..features import FeatureEdgeDetector
from ..connection import SurfaceConnection
from ...utils.argument_check import *

# 2D implementations
from .faces2d import *
from .vertex2d import *
from .curvature_vertex import PrincipalDirectionsVertices
from .curvature_faces import PrincipalDirectionsFaces

# 3D implementations
from .vertex3d import *
from .cells import *

@allowed_mesh_types(SurfaceMesh)
def SurfaceFrameField(
    mesh : SurfaceMesh,
    elements : str,
    order : int = 4,
    features : bool = True,
    verbose : bool = False,
    n_smooth : int = 3,
    smooth_attach_weight : float = None,
    use_cotan : bool = True,
    cad_correction : bool = True,
    smooth_normals : bool = True,
    singularity_indices : Attribute = None,
    custom_connection : SurfaceConnection = None,
    custom_features : FeatureEdgeDetector = None
) -> FrameField :
    """
    Computes a smooth frame field on a surface mesh.

    Args:
        mesh (SurfaceMesh): the supporting mesh onto which the framefield is based

        elements (str): "vertices" or "faces", the mesh elements onto which the frames live.
    
    Keyword Args:
        order (int, optional): Order of the frame field (number of branches). Defaults to 4.
        
        features (bool, optional): Whether to consider feature edges or not. 
            If no 'custom_features' argument is provided, features will be automatically detected (see the FeatureEdgeDetector class). Defaults to True.
        
        n_smooth (int, optional): Number of smoothing steps to perform. Defaults to 3.
        
        smooth_attach_weight (float, optional): Custom attach weight to previous solution during smoothing steps. 
            If not provided, will be estimated automatically during optimization. Defaults to None.

        use_cotan (bool, optional): whether to use cotan for a better approximation of the Laplace-Beltrami operator. 
            If False, will use a simple adjacency laplacian operator (See the _operators_ module). Defaults to True.

        cad_correction (bool, optional): Whether to modify the parallel transport as in [2] to prevent singularities to appear close to pointy areas. 
            Will overwrite any connection provided with the 'custom_connection' argument. Defaults to True.

        smooth_normals : Whether to initialize the frame field as a mean of adjacent feature edges (True), or following one of the edges (False). has no effect for frame field on faces. Defaults to True.

        verbose (bool, optional): verbose mode. Defaults to False.
        
        singularity_indices (Attribute, optional): custom singularity indices for the frame field. If provided, will use the algorithm described in [3] to get the smoothest frame field with these singularities.
            If elements is "vertices", the attribute should be indexed by the faces (where singularities appear)
            If elements is "faces", the attribute should be indexed by the vertices
            /!\\ Indices should respect the Poincarré-Hopf theorem. Defaults to None.
        
        custom_connection (SurfaceConnection, optional): custom connection object to be used for parallel transport. If not provided, a connection will be automatically computed (see SurfaceConnection class). Defaults to None.
        
        custom_feature (FeatureEdgeDetector, optional): custom feature edges to be used in frame field optimization. If not provided, feature edges will be automatically detected. If the 'features' flag is set to False, features of this object are ignored. Defaults to None.

    Raises:
        InvalidRangeArgumentError: 'order' should be >= 1
        InvalidRangeArgumentError: 'n_smooth' should be >= 0
        InvalidRangeArgumentError: 'smooth_attach_weight' should be >= 0

    Returns:
        FrameField: A framefield object with the correct specifications

    References: 
        - [1] _An Approach to Quad Meshing Based on Harmonic Cross-Valued Maps and the Ginzburg-Landau Theory_, Viertel and Osting (2018)
        
        - [2] _Frame Fields for CAD models_, Desobry et al. (2022)
        
        - [3] _Trivial Connections on Discrete Surfaces_, Crane et al. (2010)
    
    Example:
        [https://github.com/GCoiffier/mouette/blob/main/examples/framefield2D.py](https://github.com/GCoiffier/mouette/blob/main/examples/framefield2D.py)
    """

    ### Assert sanity of arguments
    check_argument("elements", elements, str, ["vertices", "faces"])
    check_argument("order", order, int)
    if order<1: raise InvalidRangeArgumentError("order",order, ">=1")  
    check_argument("n_smooth", n_smooth, int)
    if n_smooth<0: raise InvalidRangeArgumentError("n_smooth", n_smooth, ">=0")
    if smooth_attach_weight is not None and smooth_attach_weight<=0 :
        raise InvalidRangeArgumentError("smooth_attach_weight", smooth_attach_weight, ">=0")

    ### Build the correct FF class
    if elements=="vertices":
        if singularity_indices is not None:
            return TrivialConnectionVertices(mesh, singularity_indices, order, verbose, use_cotan=use_cotan, cad_correction=cad_correction, custom_connection=custom_connection, custom_features=custom_features)
        else:
            return FrameField2DVertices(mesh, order, features, verbose, n_smooth=n_smooth, smooth_attach_weight=smooth_attach_weight, 
            use_cotan=use_cotan, cad_correction=cad_correction, smooth_normals=smooth_normals, custom_connection=custom_connection, custom_features=custom_features)

    elif elements=="faces":
        if singularity_indices is not None:
            return TrivialConnectionFaces(mesh, singularity_indices, order=order, verbose=verbose, custom_connection=custom_connection, custom_features=custom_features)
        return FrameField2DFaces(mesh, order, features, verbose, n_smooth=n_smooth, smooth_attach_weight=smooth_attach_weight,use_cotan=use_cotan,custom_connection=custom_connection,custom_features=custom_features)

@allowed_mesh_types(SurfaceMesh)
def PrincipalDirections(
    mesh : SurfaceMesh, 
    elements : str,
    features : bool = True,
    verbose : bool = False,
    n_smooth : int = 1,
    smooth_attach_weight : float = None,
    patch_size : int = 3,
    curv_threshold : float = 0.01,
    complete_ff : bool = True,
    custom_features : FeatureEdgeDetector = None) -> FrameField:
    """
    Args:
        mesh (SurfaceMesh): the supporting mesh onto which the framefield is based

        elements (str): "vertices" or "faces", the mesh elements onto which the frames live.
    
    Keyword Args:
        features (bool, optional): Whether to consider feature edges or not. 
            If no 'custom_features' argument is provided, features will be automatically detected (see the FeatureEdgeDetector class). Defaults to True.

        verbose (bool, optional): verbose mode. Defaults to False.

        n_smooth (int, optional): Number of smoothing steps to perform. Defaults to 1.
        
        smooth_attach_weight (float, optional): Custom attach weight to previous solution during smoothing steps. 
            If not provided, will be estimated at 1 for vertex version and 1e-3 for faces version. Defaults to None.
        
        patch_size (int, optional): On vertices only. Radius (in nubmer of edges) of the neighboring patch to be considered to approximate the shape operator. Defaults to 3.
        
        curv_threshold (float, optional): On vertices only. Threshold on the eigenvalues of the shape operator. If smaller, eigenvectors will not be extracted (curvature is too small). Defaults to 0.01.
        
        complete_ff (bool, optional): Whether to smoothly interpolate directions on low curvature areas (< curv_threshold). Defaults to True.

        custom_features (FeatureEdgeDetector, optional): custom feature edges to be used in frame field optimization. If not provided, feature edges will be automatically detected. If the 'features' flag is set to False, features of this object are ignored. Defaults to None.

    Returns:
        Framefield : a frame field object representing the curvature directions

    References:
        - [1] https://en.wikipedia.org/wiki/Principal_curvature
        
        - [2] _Restricted Delaunay Triangulations and Normal Cycle_, Cohen-Steiner and Morvan (2003)

    Example:
        [https://github.com/GCoiffier/mouette/blob/main/examples/framefield2D.py](https://github.com/GCoiffier/mouette/blob/main/examples/framefield2D.py)
    """
    ### Assert sanity of arguments
    check_argument("elements", elements, str, ["vertices", "faces"])

    ### Build the correct FF class
    if elements=="vertices":
        return PrincipalDirectionsVertices(mesh, features, verbose,
        patch_size=patch_size,n_smooth=n_smooth, smooth_attach_weight=smooth_attach_weight, 
        curv_threshold=curv_threshold,complete_ff=complete_ff, custom_features=custom_features)
    elif elements=="faces":
        return PrincipalDirectionsFaces(mesh, features, verbose, 
        n_smooth=n_smooth, smooth_attach_weight=smooth_attach_weight, custom_features=custom_features)

@allowed_mesh_types(VolumeMesh)
def VolumeFrameField(
    mesh : VolumeMesh,
    elements : str,
    features : bool = True,
    n_smooth : int = 3,
    smooth_attach_weight : float = None, 
    verbose : bool = False,
    custom_boundary_features : FeatureEdgeDetector = None) -> FrameField:
    """
    Args:
        mesh (VolumeMesh): the supporting mesh

        elements (str): "vertices" or "cells", the mesh elements onto which the frames live.

    Keyword Args:
        features (bool, optional): Whether to consider feature edges or not. Has no effect on the cell implementation. Defaults to True.

        n_smooth (int, optional): Number of smoothing steps to perform. Defaults to 1.
        
        smooth_attach_weight (float, optional): Custom attach weight to previous solution during smoothing steps. 
            If not provided, will be estimated automatically before optimization. Defaults to None.

        verbose (bool, optional): verbose mode. Defaults to False.

        custom_boundary_features (FeatureEdgeDetector, optional): custom feature edges to be used in frame field optimization. If not provided, feature edges will be automatically detected. If the 'features' flag is set to False, features of this object are ignored. Has no effect on the cell implementation. Defaults to None.

    Returns:
        Framefield: a frame field object with the correct settings
    
    References:
        - [1] _Practical 3D frame field generation_, Ray et al. (2016)

    Example:
        [https://github.com/GCoiffier/mouette/blob/main/examples/framefield3D.py](https://github.com/GCoiffier/mouette/blob/main/examples/framefield3D.py)
    """

    check_argument("elements", elements, str, ["vertices", "cells"])
    
    if elements=="vertices":
        return FrameField3DVertices(mesh, features, verbose, 
        n_smooth=n_smooth, smooth_attach_weight=smooth_attach_weight, 
        custom_boundary_features=custom_boundary_features)
    elif elements=="cells":
        return FrameField3DCells(mesh, verbose,
        n_smooth=n_smooth, smooth_attach_weight=smooth_attach_weight)


"""
M.processing
submodule for algorithms applied on meshes
"""

from .misc import smooth
from .cutting import SingularityCutter
from .features import FeatureEdgeDetector
from .expmap import DiscreteExponentialMap

from .paths import shortest_path, shortest_path_to_border, shortest_path_to_vertex_set
from .border import extract_border_cycle, extract_border_cycle_all, extract_boundary_of_surface, extract_boundary_of_volume

from . import trees
from .parametrization import distortion
from .connection import SurfaceConnectionFaces, SurfaceConnectionVertices, FlatConnectionVertices, FlatConnectionFaces

from .point_cloud_utils import PointCloudNormalEstimator


from ..mesh.datatypes import *
from ..mesh import RawMeshData
from .connection import SurfaceConnectionVertices
from .worker import Worker
from .. import geometry as geom
from ..geometry import Vec
from ..utils import PriorityQueue

from collections.abc import Iterable
from math import pi

class DiscreteExponentialMap(Worker):
    """
    References:
        - [1] Interactive decal compositing with discrete exponential maps, Schmidt et al., 2006

    Example:
        [https://github.com/GCoiffier/mouette/blob/main/examples/expmap.py](https://github.com/GCoiffier/mouette/blob/main/examples/expmap.py)
    """
    def __init__(self, 
        mesh : SurfaceMesh, 
        conn : SurfaceConnectionVertices,
        radius : int = 2,
        verbose : bool = False
    ):
        """
        Args:
            mesh (SurfaceMesh): input surface mesh
            conn (SurfaceConnectionVertices): connection over vertices for the mesh.
            radius (int, optional): Radius in number of edges of the exponential map of each vertex. Defaults to 2.
            verbose (bool, optional): verbose options. Defaults to False.
        """
        super().__init__("ExpMap", verbose)
        self.mesh = mesh
        self.conn = conn
        self.radius = radius
        self._map : dict = dict()

    def run(self, vertices : Iterable = None):
        if vertices is None : 
            vertices = self.mesh.id_vertices
        for u in vertices:
            self._compute_map(u)
    
    def _compute_map(self, u : int):
        """
        Computes map originating from vertex u for all vertices v at most at distance 'self.d' from u in terms of number of edges
        """
        self._map[u] = dict()
        
        ### Perform graph traversal from origin u and stop as soon as distance > self.d
        visited = set()
        dist = dict()
        rot = dict()

        q = PriorityQueue()
        q.push(u,0) # weights are distance to the origin u
        dist[u] = 0
        rot[u] = 0.
        self._map[u][u] = Vec(0.,0.)

        while not q.empty():
            front = q.get()
            v,rad = front.x, front.priority
            if rad>=self.radius : continue # ignore since too far from origin
            if v in visited : continue # vertex was already set
            visited.add(v)
            for nv in self.mesh.connectivity.vertex_to_vertices(v):
                dv = dist[v] + geom.distance(self.mesh.vertices[v], self.mesh.vertices[nv])
                if dist.get(nv, float("inf")) > dv:
                    dist[nv] = dv
                    E = self.mesh.vertices[nv] - self.mesh.vertices[v]
                    E = self.conn.project(E, v)
                    rot[nv] = rot[v] + self.conn.transport(v,nv) - self.conn.transport(nv,v) + pi
                    self._map[u][nv] = self._map[u][v] + geom.rotate_2d(E, rot[v])
                if nv not in visited:
                    q.push(nv, rad+1)

    def map(self, v_source : int, v_target : int):
        """
        Returns the uv-coordinates of v_target in the tangent plane of v_source.
        If the map of v_source was not computed, it is computed on the go.

        Args:
            v_source (int): target vertex index, center of the exponential map
            v_target (int): query vertex index

        Returns:
            Vec: uv-coordinates of v_target as projected in the tangent plane of v_source. Returns None if v_target is not in the map of v_source (vertex is invalid or too far)
        """
        if v_source not in self._map:
            self._compute_map(v_source)
        return self._map[v_source].get(v_target, None)

    def export_map_as_mesh(self, v_source : int) -> SurfaceMesh:
        """
        Exports the map of vertex v_source as a flat surface mesh for visualization purposes 

        Args:
            v_source (int): source vertex index

        Returns:
            SurfaceMesh: exponential map of v_source as a surface mesh
        """
        m = RawMeshData()
        index_map = dict()

        if v_source not in self._map:
            self._compute_map(v_source)
        
        ### Add root
        m.vertices.append([0.,0.,0.])
        index_map[v_source] = 0.

        ### Compute point positions
        for iv,(v,pV) in enumerate(self._map[v_source].items()):
            index_map[v] = iv+1
            m.vertices.append(Vec(pV.x, pV.y, 0.))
        
        ### Add faces
        for F in self.mesh.faces:
            if all([x in self._map[v_source] for x in F]):
                m.faces.append([ index_map[x] for x in F ])

        return SurfaceMesh(m)





from . import Vec
import numpy as np
from ..utils.argument_check import check_argument
from .geometry import norm


class AABB:

    class IncompatibleDimensionError(Exception):
        def __init__(self, message):
            super().__init__(message)

    def __init__(self, p_min, p_max):
        """Axis Aligned Bounding Box in n dimensions.

        Args:
            p_min (Iterable): minimal values for each dimension
            p_max (Iterable): maximal values for each dimension

        Raises:
            Exception: fails if p_min and p_max have different sizes (inconsistent dimension)
        """
        self._p1 = Vec(p_min)
        self._p2 = Vec(p_max)
        if self._p1.size != self._p2.size:
            raise Exception("AABB: received two initial arrays of a different dimension!")
    
    def __repr__(self):
        return f"AABB | {self._p1} -> {self._p2}"

    @property
    def dim(self) -> int:
        """Dimension of the axis-aligned bounding box

        Returns:
            int: dimension
        """
        return self._p1.size

    @classmethod
    def unit_cube(cls, dim:int, centered:bool = False) -> "AABB":
        """Computes the unit bounding box [0,1]^n or [-0.5, 0.5]^n

        Args:
            dim (int): dimension of the BB to build
            centered (bool, optional): whether to generate [0,1]^n (False) or [-0.5;0.5]^n (True). Defaults to False.

        Returns:
            AABB: a hypercube of side length 1
        """
        if centered:
            return AABB(np.full(dim, -0.5), np.full(dim, 0.5))
        else:
            return AABB(np.zeros(dim, dtype=float), np.ones(dim, dtype=float))

    @classmethod
    def infinite(cls, dim:int) -> "AABB":
        """Computes a bounding box with bounds at infinity, containing the whole space R^n

        Args:
            dim (int): dimension of the BB to build

        Returns:
            AABB: A bounding box containing all of R^n
        """
        return AABB(np.full(dim, -np.inf), np.full(dim, np.inf))

    @classmethod
    def of_points(cls, points: np.ndarray, padding: float = 0.) -> "AABB":
        """Computes the bounding box of a set of nD points.

        Args:
            points (np.ndarray): input points
            padding (float, optional): slack to be added between the mesh and the box. Defaults to 0 for a tight bounding box.

        Returns:
            AABB: nD bounding box of the points
        """
        points = np.array(points)
        if len(points.shape) != 2:
            raise Exception("Expected an array of points of size (N,d) with N being the number of points and d their dimension")
        pad = np.full(points.shape[1], padding)
        pt_min = np.min(points, axis=0) - pad
        pt_max = np.max(points, axis=0) + pad
        return AABB(pt_min, pt_max)

    @classmethod
    def of_mesh(cls, mesh: "Mesh", padding: float = 0.) -> "AABB":
        """Computes the 3D bounding box of all vertices of a mesh

        Args:
            mesh (Mesh): input mesh. Can be of any type (only the 'vertices' container is accessed)
            padding (float, optional): slack to be added between the mesh and the box. Defaults to 0 for a tight bounding box.

        Returns:
            AABB: 3D bounding box of the vertices of the mesh
        """
        pad = Vec(padding, padding, padding)
        pt_min = np.min(mesh.vertices._data, axis=0) - pad
        pt_max = np.max(mesh.vertices._data, axis=0) + pad
        return AABB(pt_min, pt_max)

    @property
    def mini(self) -> Vec:
        """Minimum coordinates of any point inside the box
        The box is an axis-aligned hexahedron which opposite corners are mini and maxi

        Returns:
            Vec: minimum coordinates
        """
        return self._p1
    
    @property
    def maxi(self) -> Vec:
        """Maximum coordinates of any point inside the box. 
        The box is an axis-aligned hexahedron which opposite corners are mini and maxi

        Returns:
            Vec: maximum coordinates
        """
        return self._p2
    
    @property
    def span(self) -> Vec:
        """Dimensions of the box

        Returns:
            Vec: dimensions of the box
        """
        return self._p2 - self._p1

    @property
    def center(self) -> Vec:
        """Coordinates of the center point

        Returns:
            Vec: center point
        """
        return (self._p1 + self._p2)/2

    @staticmethod
    def intersection(b1: "AABB", b2: "AABB") -> "AABB":
        """
        Computes the intersection bounding box of two bounding boxes

        Args:
            b1 (AABB): first bounding box
            b2 (AABB): second bounding box

        Raises:
            AABB.IncompatibleDimensionError: fails if b1 and b2 have different dimensions

        Returns:
            AABB: a bounding box representing the intersection (may be empty).
        """
        if b1.dim != b2.dim: 
            raise AABB.IncompatibleDimensionError(f"Bounding boxes have different dimensions ({b1.dim} and {b2.dim}) : intersection impossible")
        
        return AABB(np.maximum(b1.mini, b2.mini), np.minimum(b1.maxi, b2.maxi))
    
    def __and__(self,other):
        return AABB.intersection(self,other)

    @staticmethod
    def do_intersect(b1: "AABB", b2: "AABB") -> bool:
        """
        Intersection test between two bounding boxes

        Args:
            b1 (AABB): first bounding box
            b2 (AABB): second bounding box

        Raises:
            AABB.IncompatibleDimensionError: fails if b1 and b2 have different dimensions
            
        Returns:
            bool: whether the two BB intersect
        """
        if b1.dim != b2.dim: 
            raise AABB.IncompatibleDimensionError(f"Bounding boxes have different dimensions ({b1.dim} and {b2.dim}): intersection impossible")
    
        predicates = [b1.mini[i] <= b2.maxi[i] and b1.maxi[i] >= b2.mini[i] for i in range(b1.dim)]
        return np.all(predicates)
    
    @staticmethod
    def union(b1: "AABB", b2: "AABB") -> "AABB":
        """Computes the union bounding box of two bounding boxes

        Args:
            b1 (AABB): first bounding box
            b2 (AABB): second bounding box

        Raises:
            AABB.IncompatibleDimensionError: fails if b1 and b2 have different dimensions
        
        Returns:
            AABB: a bounding box representing the union
        """
        if b1.dim != b2.dim: 
            raise AABB.IncompatibleDimensionError(f"Bounding boxes have different dimensions ({b1.dim} and {b2.dim}): union impossible")
        pmin = np.min((b1._p1, b2._p1), axis=0)
        pmax = np.max((b1._p2, b2._p2), axis=0)
        return AABB(pmin, pmax)
    
    def __or__(self, other):
        return AABB.union(self,other)

    def pad(self, pad:Vec):
        """Enlarges the bounding box by adding `pad` on each dimension on each side. Does nothing if the padding values are negative.

        Args:
            pad (float | iterable): Additional dimensions to be added. If a float is provided, will assume that padding is the same for each dimensions. If an array is provided, its size should match the dimension of the box. Values are clamped to be >=0.
        """
        if isinstance(pad,float):
            pad = np.full(self.dim, pad)
        else:
            pad = Vec(pad)
            if pad.size != self.dim: raise AABB.IncompatibleDimensionError(f"Padding vector has a different dimension ({pad.size}) than bounding box ({self.dim})") 
        pad = np.maximum(pad, 0)
        self._p1 -= pad
        self._p2 += pad

    def contains_point(self, pt: Vec) -> bool:
        """Point - bounding box intersection predicate.
        If the point is on the boundary of the box, the convention is as follows: inclusive if the point touches the min coordinates, exclusive for the max coordinates

        Args:
            pt (Vec): a query position
        
        Raises:
            AABB.IncompatibleDimensionError: fails if the point has a different dimension than the bounding box.

        Returns:
            bool: whether the point 'pt' is inside the bounding box.
        """
        pt = Vec(pt)
        if pt.size != self.dim:
            raise AABB.IncompatibleDimensionError(f"Point has a different dimension({pt.size}) than bounding box ({self.dim})")
        return (self._p1 <= pt).all() and (pt < self._p2).all()
    
    def project(self, pt:Vec) -> Vec:
        """Computes the closest point from point 'pt' in the bounding box in Euclidean distance

        Args:
            pt (Vec): the point to project
        
        Raises:
            AABB.IncompatibleDimensionError: fails if the point has a different dimension than the bounding box.

        Returns:
            Vec: the projected point
        """
        pt = Vec(pt)
        if pt.size != self.dim:
            raise AABB.IncompatibleDimensionError(f"Point has a different dimension({pt.size}) than bounding box ({self.dim})")
        if self.contains_point(pt): return pt # pt is its own projection
        return np.maximum(self.mini, np.minimum(self.maxi, pt))

    def distance(self, pt:Vec, which="l2") -> float:
        """Computes the distance from a point to the bounding box.

        Args:
            pt (Vec): coordinates of the point
            which (str, optional): which distance to consider. Choices are 'l2', 'l1' or 'linf'. Defaults to "l2".

        Raises:
            AABB.IncompatibleDimensionError: fails if the point has a different dimension than the bounding box.
            
        Returns:
            float: the distance from the point to the bounding box
        """
        pt = Vec(pt)
        if pt.size != self.dim:
            raise AABB.IncompatibleDimensionError(f"Point has a different dimension({pt.size}) than bounding box ({self.dim})")
        check_argument("which", which, str, ["l2", "l1", "linf"])
        vec = np.maximum(np.maximum(self.mini - pt, pt - self.maxi), 0.)
        return norm(vec,which)

    def is_empty(self)->bool:
        """Tests if the bounding box encloses an empty domain

        Returns:
            bool: whether the bounding box is empty
        """
        return np.any(self.mini >= self.maxi)


from functools import wraps

from .base import Mesh
from .pointcloud import PointCloud
from .linear import PolyLine
from .surface import SurfaceMesh
from .volume import VolumeMesh

class BadMeshTypeException(Exception):
    def __init__(self, typ, message):
        self.type = typ
        super().__init__(message)

def type_to_str(typ):
    if typ==PointCloud: return "mouette.PointCloud"    
    if typ==PolyLine: return "mouette.PolyLine"    
    if typ==SurfaceMesh: return "mouette.SurfaceMesh"    
    if typ==VolumeMesh: return "mouette.VolumeMesh"
    raise BadMeshTypeException(typ, "[Mouette] Unknown mesh type")

def allowed_mesh_types(*allowed_types : list):
    def decorator(function):
        @wraps(function)
        def wrapper(*args, **kwargs):
            for arg in args:
                if isinstance(arg, Mesh) and type(arg) not in filter(lambda t : issubclass(t,Mesh), allowed_types):
                    raise BadMeshTypeException(type(arg), "[Mouette] Mesh type '{}' is not allowed for this function. Allowed types : {}".format( type_to_str(type(arg)), [type_to_str(u) for u in allowed_types]))
            result = function(*args, **kwargs)
            return result
        return wrapper
    return decorator

def forbidden_mesh_types(*forbidden_types : list):
    def decorator(function):
        @wraps(function)
        def wrapper(*args, **kwargs):
            for arg in args:
                if isinstance(arg, Mesh) and type(arg) in filter(lambda t : issubclass(t,Mesh), forbidden_types):
                    raise BadMeshTypeException("[Mouette] Mesh type '{}' is forbidden for this function. Forbidden types are {}".format( type_to_str(type(arg)), [type_to_str(u) for u in forbidden_types]))
            result = function(*args, **kwargs)
            return result
        return wrapper
    return decorator


from .argument_check import *

def get_filename(filepath : str) -> str:
    file = filepath.split("/")[-1]
    return file.split(".")[0]

def get_extension(filepath : str) -> str:
    return filepath.split(".")[-1]

def replace_extension(path, new_extension) -> str:
    return ".".join(path.split(".")[:-1]) + new_extension

def keyify(*args):
    if len(args)==1:
        key = [x for x in args[0]]
    else :
        key = [x for x in args]
    key.sort()
    return tuple(key)

def replace_in_list(l : list, x, y) -> list:
    """
    Returns the list where all occurences of x have been replaced by value y

    Args:
        l (list): list of elements
        x (any): value to be replaced
        y (any): value to replace

    Returns:
        list: the list where all occurences of x bave been replaced by value y
    """
    return [y if e==x else e for e in l]

class Logger:
    def __init__(self, name = "Logger", verbose=True):
        self.name = name
        self.verbose = verbose

    def log(self, *messages):
        if self.verbose:
            print(*((f"[{self.name}]",) + messages))


from .worker import Worker

from ..geometry import distance
from ..mesh.datatypes import *
from ..mesh.mesh_data import RawMeshData
from ..mesh.mesh_attributes import *
from ..utils import keyify, UnionFind, PriorityQueue
from .paths import shortest_path, shortest_path_to_border, shortest_path_to_vertex_set
from .trees import FaceSpanningForest
from .. import attributes
from collections import deque

class SingularityCutter(Worker):
    """
    Given some indexes in the mesh, performs optimal cuts connecting all these vertices with the boundary.
    """

    @allowed_mesh_types(SurfaceMesh)
    def __init__(self, 
            mesh : SurfaceMesh,
            singularities : list,
            features : "FeatureEdgeDetector" = None,
            verbose = False):
        """
        Args:
            mesh (SurfaceMesh): input mesh
            singularities (list): indices of the singular vertices
            features (FeatureEdgeDetector, optional): feature edge data structure. If provided, the cuts will follow the feature as much as possible. Defaults to None.
            verbose (bool, optional): verbose mode. Defaults to False.

        Attributes:
            cut_edges (set): indices of edges that were cut
            cut_adj (dict): cut graph given as adjacency lists per vertex
        """
        super().__init__("SingularityCutter", verbose)
        self.input_mesh : SurfaceMesh = mesh
        
        if isinstance(singularities, list):
            self.singularities = singularities
            self.singu_set = set(singularities)
        else:
            self.singularities : list = [_x for _x in singularities]
            self.singu_set = set(singularities)

        self.edge_lengths = attributes.edge_length(self.input_mesh, persistent=False)

        self.feat_detector : "FeatureEdgeDetector" = features
        self._has_features : bool = None

        self.cut_edges : set = None # ids of edges of input mesh that were cut 
        self.cut_adj : dict = None # vertex -> set of adjacent vertices in the cut graph

        # /!\ global ordering of vertices/edges is likely to change. Only triangles remain the same
        self.ref_vertex : dict = None # split vertex -> original vertex

        self._cut_graph : PolyLine = None
        self._output_mesh : SurfaceMesh = None
    
    @property
    def has_features(self) -> bool:
        """Whether the input mesh has feature edges defined"""
        if self._has_features is None:
            self._has_features = False
            if self.feat_detector is not None:
                for e in self.feat_detector.feature_edges:
                    if not self.input_mesh.is_edge_on_border(*self.input_mesh.edges[e]):
                        self._has_features = True
                        break
        return self._has_features

    @property
    def output_mesh(self) -> SurfaceMesh:
        """
        Returns:
            SurfaceMesh: the mesh where cuts have been performed
        """
        if self._output_mesh is None:
            self._build_mesh_with_cuts()
        return self._output_mesh
    
    @property
    def cut_graph(self) -> PolyLine:
        """
        Returns:
            PolyLine: the cut edges as a Polyline
        """
        if self._cut_graph is None:
            self._build_cut_graph_as_mesh()
        return self._cut_graph

    def run(self):
        """
        Runs the cutting process
        """
        self.log("Cutting to link singularities and retrieve disk topology")
        self.log("# Singularities :", len(self.singularities))
        if self.has_features:
            self.log(f"{len(self.feat_detector.feature_edges)} feature edges and {len(self.feat_detector.feature_vertices)} feature vertices provided")
            self._run_with_features()
        else:
            self._run_no_features()

    def _run_with_features(self):
        self.log("Step 1 : Minimal Spanning Tree to link singularities and border")
        edge_flag = self._build_singularity_spanning_tree_with_features()

        self.log("Step 2 : BFS on dual graph + BFS on non-traversed edges to extract cuts and homology")
        evisited = self._build_dual_tree_with_features(edge_flag)
        self._build_cut_edges_tree(evisited) # builds attributes self.cut_edges and self.cut_adj

        self.log("Step 3 : pruning homology tree")
        self._prune_edge_tree()
        self.log("Cutting done")

    def _run_no_features(self):
        
        self.log("Step 1 : Minimal Spanning Tree to link singularities and border")
        edge_flag = self._build_singularity_spanning_tree_no_features()
    
        self.log("Step 2 : BFS on dual graph + BFS on non-traversed edges to extract cuts and homology")
        evisited = self._build_dual_tree_no_features(edge_flag)
        self._build_cut_edges_tree(evisited) # builds attributes self.cut_edges and self.cut_adj

        self.log("Step 3 : pruning homology tree")
        self._prune_edge_tree()
        self.log("Cutting done")

    def _build_singularity_spanning_tree_no_features(self):
        """
        Returns:
            Attribute: the spanning tree given as a boolean attribute on edges
        """
        mesh_has_border = len(self.input_mesh.boundary_vertices)>0

        BORDER = -1 # the index of border (>0 are vertices)
        singul = self.singularities + [BORDER] if mesh_has_border else self.singularities

        # edge_flags = self.input_mesh.edges.create_attribute("singularity_tree", bool) 
        edge_flags = Attribute(bool) # edge selected for spanning tree
        
        if not self.singularities:
            # no singularities => no spanning tree and no constraints on edges
            return edge_flags
    
        # Compute all edge paths between singularities (and border)
        path_btw_singus = dict()
        for i,a in enumerate(self.singularities):
            paths_a = shortest_path(self.input_mesh, a, set(self.singularities[i:]), weights=self.edge_lengths)
            for b in paths_a:
                path_btw_singus[keyify(a,b)] = paths_a[b]
            if mesh_has_border:
                path_btw_singus[(BORDER,a)] = shortest_path_to_border(self.input_mesh, a, weights=self.edge_lengths)

        # Compute path lengths
        def compute_path_length(path): # utility function for readability
            l = 0
            for i in range(1, len(path)):
                A,B = path[i-1], path[i]
                l += self.edge_lengths[self.input_mesh.connectivity.edge_id(A,B)]
            return l
        
        path_lengths = []
        for k in path_btw_singus:
            path_lengths.append((compute_path_length(path_btw_singus[k]), k))
        path_lengths.sort()

        # Build Minimum Spanning Tree using Kruskal's algorithm
        uf = UnionFind(singul)
        selected = []
        for (_,key) in path_lengths: # first elem in tuple is length (not useful here)
            a,b = key
            if not uf.connected(a,b):
                selected.append(key)
                uf.union(a,b)

        # Flag edges
        for (a,b) in selected:
            path_ab = path_btw_singus[(a,b)]
            for i in range(1, len(path_ab)):
                u,v = path_ab[i-1], path_ab[i]
                edge_flags[ self.input_mesh.connectivity.edge_id(u,v) ] = True
        return edge_flags

    def _build_singularity_spanning_tree_with_features(self):
        """We do not apply the same algorithm as we want the spanning tree to have the maximal possible intersection with the feature graph. This leads to prettier results.

        Returns:
            Attribute: the spanning tree given as a boolean attribute on edges
        """
        edge_flags = self.input_mesh.edges.create_attribute("singularity_tree", bool) # edge selected for spanning tree
        if len(self.singularities)==0:
            # no singularities => no spanning tree and no constraints on edges
            return edge_flags

        # First compute the closest point from singularities to feature graph
        closest_v = set()
        for v in self.singularities:
            id_feat, path = shortest_path_to_vertex_set(self.input_mesh, v, self.feat_detector.feature_vertices, weights=self.edge_lengths)
            closest_v.add(id_feat)
            for i in range(len(path)-1):
                u,v = path[i], path[i+1]
                e = self.input_mesh.connectivity.edge_id(u,v)
                edge_flags[e] = True

        # Then construct a tree on the feature graph linking all the previous points. Perform a BFS
        closest_v = list(closest_v)
        queue = deque()
        visited = dict([(v, False) for v in self.feat_detector.feature_vertices])
        parent = dict([(v, None) for v in self.feat_detector.feature_vertices])
        for v in closest_v:
            queue.append((v,None))
        while len(queue)>0:
            v,prev = queue.popleft()
            if visited[v] : continue
            visited[v] = True
            parent[v] = prev
            if prev is not None:
                e = self.input_mesh.connectivity.edge_id(v,prev)
                edge_flags[e] = True
            for e in self.input_mesh.connectivity.vertex_to_edges(v):
                if e in self.feat_detector.feature_edges:
                    nv = self.input_mesh.connectivity.other_edge_end(e,v)
                    if not visited[nv]:
                        queue.append((nv,v))
        return edge_flags

    def _build_feature_regions(self, forbidden_edges):
        if not self.has_features: return None
        regions = UnionFind(self.input_mesh.id_faces)
        not_traversible = { _e for _e in forbidden_edges} | self.feat_detector.feature_edges
        tree = FaceSpanningForest(self.input_mesh, not_traversible)()
        for vertex,father in tree.traverse():
            if father is not None:
                regions.union(vertex,father)

        TriFlagAttr = ArrayAttribute(int, len(self.input_mesh.faces)) # self.input_mesh.faces.create_attribute("face_regions", int)
        for f in self.input_mesh.id_faces:
            TriFlagAttr[f] = regions.find(f)
        return regions

    def _build_dual_tree_no_features(self, forbidden_edges:Attribute):
        # Dijsktra on faces (dual edges)
        fvisited = ArrayAttribute(bool, len(self.input_mesh.faces)) #self.input_mesh.faces.create_attribute("cut_visited", bool)
        path = [None for _ in self.input_mesh.id_faces]
        dist = [float("inf") for _ in self.input_mesh.id_faces]
        queue : PriorityQueue = PriorityQueue() # queue contains indexes of faces
        queue.push(0,0)
        dist[0] = 0

        barycenters = attributes.face_barycenter(self.input_mesh, persistent=False)

        def face_distance(f1,f2):
            # Heuristic for distance between two faces
            return distance(barycenters[f1], barycenters[f2])

        while not queue.empty():
            iF = queue.get().x
            if fvisited[iF] : continue
            fvisited[iF] = True                
            for e in self.input_mesh.connectivity.face_to_edges(iF):
                v1,v2 = self.input_mesh.edges[e]
                if forbidden_edges[e] : continue # edge is on the singularity spanning tree
                iF2 = self.input_mesh.connectivity.opposite_face(v1,v2,iF)
                if iF2 is not None: 
                    d = face_distance(iF,iF2)
                    if dist[iF2] > dist[iF] + d :
                        dist[iF2] = dist[iF] + d
                        path[iF2] = e
                    if not fvisited[iF2] :
                        queue.push(iF2, dist[iF2])
        return {path[f] for f in self.input_mesh.id_faces if path[f] is not None }
    
    def _build_dual_tree_with_features(self, forbidden_edges:Attribute):
        regions = self._build_feature_regions(forbidden_edges) # face regions delimited by feature edges

        # Dijsktra on faces (dual edges)
        fvisited = ArrayAttribute(bool, len(self.input_mesh.faces)) #self.input_mesh.faces.create_attribute("cut_visited", bool)
        path = [None for _ in self.input_mesh.id_faces]
        dist = [float("inf") for _ in self.input_mesh.id_faces]
        queue : PriorityQueue = PriorityQueue() # queue contains indexes of faces
        queue.push(0,0)
        dist[0] = 0

        barycenters = attributes.face_barycenter(self.input_mesh, persistent=False)

        def face_distance(f1,f2):
            # Heuristic for distance between two faces
            return distance(barycenters[f1], barycenters[f2])

        while not queue.empty():
            iF = queue.get().x
            if fvisited[iF] : continue
            fvisited[iF] = True        
            for e in self.input_mesh.connectivity.face_to_edges(iF):
                v1,v2 = self.input_mesh.edges[e]
                if forbidden_edges[e] : continue # edge is on the singularity spanning tree
                iF2 = self.input_mesh.connectivity.opposite_face(v1,v2,iF)
                if iF2 is not None:
                    u,v = set(self.input_mesh.faces[iF]) & set(self.input_mesh.faces[iF2])
                    blocked = self.input_mesh.connectivity.edge_id(u,v) in self.feat_detector.feature_edges and regions.connected(iF,iF2)
                    if not blocked:
                        d = face_distance(iF,iF2)
                        if dist[iF2] > dist[iF] + d :
                            dist[iF2] = dist[iF] + d
                            path[iF2] = e
                        if not fvisited[iF2]:
                            queue.push(iF2, dist[iF2])
                            regions.union(iF,iF2)
        return {path[f] for f in self.input_mesh.id_faces if path[f] is not None }

    def _build_cut_edges_tree(self, evisited):
        # Build self._cut_edges with edges that were **not** used during the above search
        self.cut_edges = set(self.input_mesh.id_edges) - evisited
        # build connectivity on the edge tree
        self.cut_adj = dict([(i,set()) for i in self.input_mesh.id_vertices]) # connectivity lists of cut tree
        for e in self.cut_edges:
            a,b = self.input_mesh.edges[e]
            self.cut_adj[a].add(b)
            self.cut_adj[b].add(a)

    def _prune_edge_tree(self):
        # remove leaves of cut tree until we fall on singularities
        queue = deque()
        for i in self.input_mesh.id_vertices:
            d = len(self.cut_adj[i])
            if d==1 and i not in self.singularities:
                queue.append(i)
        while len(queue)>0:
            A = queue.popleft()
            for B in self.cut_adj[A]:
                self.cut_adj[B].remove(A)
                self.cut_edges.remove(self.input_mesh.connectivity.edge_id(A,B))
                if len(self.cut_adj[B])==1 and B not in self.singularities:
                    queue.append(B)
            self.cut_adj[A] = set() # A is disconnected
        self.log("# Edges cut:", len(self.cut_edges) - len(self.input_mesh.boundary_edges))

    def _build_cut_graph_as_mesh(self):
        # Build tree of potential cuts for debug purposes
        self._cut_graph = RawMeshData()
        hard_edges = self._cut_graph.edges.create_attribute("hard_edges", bool)

        new_v_id = dict()
        vid = 0
        for ie in self.cut_edges:
            v1,v2 = self.input_mesh.edges[ie]
            for v in (v1,v2):
                if v not in new_v_id:
                    new_v_id[v] = vid
                    vid += 1
                    self._cut_graph.vertices.append(self.input_mesh.vertices[v])
            e = keyify(new_v_id[v1], new_v_id[v2])
            hard_edges[len(hard_edges)] = True
            self._cut_graph.edges.append(e)

        singuls_attr = self._cut_graph.vertices.create_attribute("selection", bool)
        for x in self.singularities:
            singuls_attr[new_v_id[x]] = True
        self._cut_graph = PolyLine(self._cut_graph)

    def _build_mesh_with_cuts(self):
        self._output_mesh = RawMeshData()
        uf = UnionFind(range(3*len(self.input_mesh.faces)))
        duplicate_vertices = dict([(v, set()) for v in self.input_mesh.id_vertices])

        # At first, no faces are adjacent
        kF = 0
        for iF,F in enumerate(self.input_mesh.faces):
            nF = len(F)
            self._output_mesh.faces.append([kF+_i for _i in range(nF)])
            for iv,v in enumerate(F):
                pv = self.input_mesh.vertices[v]
                self._output_mesh.vertices.append(pv)
                duplicate_vertices[v].add(kF+iv)
            kF += nF

        # Fusion vertices along edges that are not a cut
        for e in self.input_mesh.interior_edges:
            a,b = self.input_mesh.edges[e]
            if e not in self.cut_edges:
                F1, iA1, iB1 = self.input_mesh.connectivity.direct_face(a,b,True)
                F2, iB2, iA2 = self.input_mesh.connectivity.direct_face(b,a,True)
                uf.union(self._output_mesh.faces[F1][iA1], self._output_mesh.faces[F2][iA2])
                uf.union(self._output_mesh.faces[F1][iB1], self._output_mesh.faces[F2][iB2])

        for i,F in enumerate(self._output_mesh.faces):
            self._output_mesh.faces[i] = [uf.find(v) for v in F]
            
        # Reorder vertices to [0,n-1] and get rid of non connected duplicates
        imap = dict()
        i=0
        for F in self._output_mesh.faces:
            for v in F:
                if v in imap: continue
                imap[v]=i # assign new index to first version of vertex u met
                i+=1

        for iF,F in enumerate(self._output_mesh.faces):
            self._output_mesh.faces[iF] = [imap[v] for v in F]
        order_verts = [None]*len(imap)

        for u in range(len(self._output_mesh.vertices)):
            if u not in imap: continue
            order_verts[imap[u]] = self._output_mesh.vertices[u]
        self._output_mesh.vertices.clear()
        self._output_mesh.vertices += order_verts
        self._output_mesh = SurfaceMesh(self._output_mesh)
        
        # get rid of duplicated and isolated vertices in map
        for v in duplicate_vertices:
            duplicate_vertices[v] = {imap[uf.find(u)] for u in duplicate_vertices[v]}

        # inverse duplicate_vertices to get self.ref_vertex
        self.ref_vertex = dict()
        for v in duplicate_vertices:
            for u in duplicate_vertices[v]:
                self.ref_vertex[u] = v




import numpy as np
from ..mesh.datatypes import *
from ..mesh.mesh import from_arrays, merge
from ..mesh.mesh_data import RawMeshData, DataContainer
from ..attributes.misc_vertices import vertex_normals
from ..geometry import Vec
from .. import utils
from .bezier import BezierCurve

from abc import ABC, abstractmethod

class _PnTriangulation(ABC):
    def __init__(self, nodes : np.ndarray, faces : np.ndarray, order: int):
        self.nodes = DataContainer(nodes, id="nodes")
        self.faces = DataContainer(faces, id="faces")
        self._order : int = order

    @classmethod
    def from_P1_mesh(cls, mesh: SurfaceMesh):
        return
    
    @abstractmethod
    def rasterize(self, res: int = 10):
        return
    
    @property
    def id_nodes(self):
        return range(len(self.nodes))
    
    @property
    def id_faces(self):
        return range(len(self.nodes))

    @property
    def dim(self) -> int:
        return self.nodes.shape[1]
    
    @property
    def order(self) -> int:
        return self._order

    @property
    def control_points(self) -> PointCloud:
        pc_nodes = from_arrays(np.array(self.nodes))
        for attr_name in self.nodes.attributes:
            pc_nodes.vertices._attr[attr_name] = self.nodes.get_attribute(attr_name)
        return pc_nodes

def rasterize_pn_as_polylines(nodes, faces, pattern, res: int = 10):
    polylines = []
    visited = set()
    for F in faces:
        for E in pattern:
            EF = [F[e] for e in E]
            if (k := utils.keyify(EF)) in visited: continue
            line = BezierCurve([nodes[_i] for _i in EF]).as_polyline(res)
            polylines.append(line)
            visited.add(k)
    return merge(polylines)

###### ORDER 2 ######

def evaluate_P2(A,B,C,AB,BC,CA, u,v):
    l1,l2,l3 = 1-u-v, u, v
    return l1*l1*A + l2*l2*B + l3*l3*C + 2*l1*l2*AB + 2*l2*l3*BC + 2*l1*l3*CA

class P2Triangulation(_PnTriangulation):

    def __init__(self, nodes: np.ndarray, faces: np.ndarray):
        """
        v
        ^
        |
        2
        |`\ 
        |  `\ 
        5    `4 
        |      `\ 
        |        `\ 
        0-----3----1  -> u


        Args:
            nodes (np.ndarray): array of size V*3 storing nodes' coordinates.
            faces (np.ndarray): array of size F*6 storing face indices of the mesh. Faces are indexes using gmsh convention (see diagram above)
        """
        super().__init__(nodes, faces, 2)
        self.connectivity = P2Triangulation._Connectivity(self)

    @property
    def node_order(self):
        if not self.nodes.has_attribute("order"):
            # Compute node order
            node_order = self.nodes.create_attribute("order",int, dense=True)
            for A0,A1,A2,A3,A4,A5 in self.faces:
                node_order[A0] = 1
                node_order[A1] = 1
                node_order[A2] = 1
                node_order[A3] = 2
                node_order[A4] = 2
                node_order[A5] = 2
        return self.nodes.get_attribute("order")

    @classmethod
    def from_P1_mesh(cls, mesh: SurfaceMesh):
        """Generates an order 2 mesh from a simple (first order) triangular mesh.

        Splits all edges into two and creates flat 2nd order Bezier triangles for each face.

        Args:
            mesh (SurfaceMesh): input triangle mesh

        Returns:
            P2Triangulation: a flat order 2 Bezier mesh.
        """
        nV = len(mesh.vertices)
        V = list(mesh.vertices)
        F = []
        for A,B in mesh.edges:
            pA, pB = V[A], V[B]
            pAB = (pA+pB)/2
            V.append(pAB)
        for A,B,C in mesh.faces:
            eAB = mesh.connectivity.edge_id(A,B)
            eBC = mesh.connectivity.edge_id(B,C)
            eCA = mesh.connectivity.edge_id(C,A)
            F.append((A,B,C, nV+eAB, nV+eBC, nV+eCA))
        return cls(np.array(V),np.array(F))

    def rasterize(self, res:int = 10, only_curves: bool = False):
        """
        Rasterize a P2 mesh as a P1 mesh for lazy rendering 
            
        Args:
            V (np.ndarray): Array of vertex positions (size Nx3)
            F (np.ndarray): Array of face indices (size Mx6). Faces are indexed in order (A,B,C,AB,BC,CA)
            res (int, optional): Resolution of each P2 element. Defaults to 10.
            only_curves (bool, optional): if set to True, will only rasterize edges as Bezier curve and not draw the surface. Defaults to False.

        Returns:
            SurfaceMesh | Polyline: a mesh approximating the P2 mesh
        """
        if only_curves:
            return rasterize_pn_as_polylines(self.nodes, self.faces, [(0,3,1), (1,4,2), (2,5,0)], res)
        else:
            return self._rasterize_as_surface(res)
        
    
    def _rasterize_as_surface(self, res=10):
        out = RawMeshData()

        def ij_to_p(i,j):
            return i*res - i*(i-1)//2 + j

        for triP2 in self.faces:
            pA,pB,pC,pAB,pBC,pCA = (self.nodes[_x] for _x in triP2)
            paramU = np.linspace(0,1,res)
            paramV = np.linspace(0,1,res)
            f_ind = len(out.vertices)

            # evaluate vertices
            for i in range(res):
                for j in range(res-i):
                    pt = Vec(evaluate_P2(pA,pB,pC,pAB,pBC,pCA, paramU[i], paramV[j]))
                    if pt.size==2:
                        pt = Vec(pt.x, pt.y, 0.)
                    out.vertices.append(pt)
                    
            # add faces
            for i in range(res-1):
                for j in range(res-1-i):
                    out.faces.append([f_ind+x for x in (ij_to_p(i,j), ij_to_p(i+1,j), ij_to_p(i,j+1))])
                    if j<res-2-i:
                        out.faces.append([f_ind+x for x in (ij_to_p(i+1,j), ij_to_p(i+1,j+1), ij_to_p(i,j+1))])
        return SurfaceMesh(out)
 
    class _Connectivity:

        def __init__(self, mesh):
            self.master = mesh
            self._connP1 : dict = None
            self._connP2 : dict = None
            self._middle : dict = None
        
        def _build(self):
            self._connP1 = dict()
            self._connP2 = dict()
            self._middle = dict()

            for A0,A1,A2,A3,A4,A5 in self.master.faces:
                self._connP1[A0] = self._connP1.get(A0,set()).union({A1,A2})
                self._connP1[A1] = self._connP1.get(A1,set()).union({A0,A2})
                self._connP1[A2] = self._connP1.get(A2,set()).union({A0,A1})

                self._middle[utils.keyify(A0,A1)] = A3
                self._middle[utils.keyify(A1,A2)] = A4
                self._middle[utils.keyify(A2,A0)] = A5
                
                self._connP2[A0] = self._connP2.get(A0, set()).union({A3,A5})
                self._connP2[A1] = self._connP2.get(A1, set()).union({A3,A4})
                self._connP2[A2] = self._connP2.get(A2, set()).union({A4,A5})
                self._connP2[A3] = [A0,A1]
                self._connP2[A4] = [A1,A2]
                self._connP2[A5] = [A0,A2]

        def vertex_to_vP1(self,v):
            if self._connP1 is None: self._build()
            return self._connP1.get(v, [])
        
        def vertex_to_vP2(self,v):
            if self._connP2 is None: self._build()
            return self._connP2.get(v,[])
        
        def middle_of_edge(self,a,b):
            if self._middle is None: self._build()
            return self._middle.get(utils.keyify(a,b),None)
        

###### ORDER 3 ######

def evaluate_P3(A,B,C,AAB, ABB, BBC, BCC, CCA, CAA, ABC, u,v):
    l1,l2,l3 = 1-u-v, u, v
    res  = l1*l1*l1*A + l2*l2*l2*B + l3*l3*l3*C
    res += 3*l1*l1*l2*AAB + 3*l2*l2*l3*BBC + 3*l3*l1*l1*CAA
    res += 3*l1*l2*l2*ABB + 3*l3*l3*l2*BCC + 3*l3*l3*l1*CCA
    res += 6*l1*l2*l3*ABC
    return res

class P3Triangulation(_PnTriangulation):

    def __init__(self, nodes: np.ndarray, faces: np.ndarray):
        """
        v
        ^
        |
        2
        | \ 
        7   6
        |     \ 
        8   9   5
        |         \ 
        0---3---4---1 -> u


        Args:
            nodes (np.ndarray): array of size V*3 storing nodes' coordinates.
            faces (np.ndarray): array of size F*10 storing face indices of the mesh. Faces are indexes using gmsh convention (see diagram above)
        """
        super().__init__(nodes, faces, 3)

    
    @classmethod
    def from_P1_mesh(cls, mesh: SurfaceMesh, curve: bool = True):
        """Generates an order 3 mesh from a simple (first order) triangular mesh.

        Args:
            mesh (SurfaceMesh): input triangular mesh (order 1)
            curve (bool, optional): if True, will apply algorithm from [1] to smooth the surface. Otherwise, will simply keep faces flat. Defaults to True.

        References:
            [1] Curved PN triangles, Vlachos et al. (2001)

        Returns:
            P3Triangulation: An order 3 Bezier triangulation
        """
                
        V = [p for p in mesh.vertices]
        F = []
        sep1 = len(mesh.vertices)
        inds = dict()
        for ie,(A,B) in enumerate(mesh.edges):
            pA, pB = mesh.vertices[A], mesh.vertices[B]
            pAAB = (2*pA+pB)/3
            pABB = (pA+2*pB)/3
            inds[(A,B)] = sep1+2*ie
            inds[(B,A)] = sep1+2*ie+1
            V += [pAAB,pABB]
        sep2 = len(V)
        for iF,(A,B,C) in enumerate(mesh.faces):
            pA, pB, pC = mesh.vertices[A], mesh.vertices[B], mesh.vertices[C]
            F.append((A,B,C, 
                inds[(A,B)], inds[(B,A)], 
                inds[(B,C)], inds[(C,B)], 
                inds[(C,A)], inds[(A,C)], 
            sep2+iF))
            V.append((pA+pB+pC)/3)

        V,F = np.array(V), np.array(F)
        if not curve: return cls(V,F)
        normals = vertex_normals(mesh, persistent=False)
        visited = np.zeros(V.shape[0], dtype=bool)
        for iA,iB,iC,iAAB,iABB, iBBC, iBCC, iCCA, iCAA, iABC in F:
            pA,pB,pC = V[iA], V[iB], V[iC]
            nA,nB,nC = normals[iA], normals[iB], normals[iC]
            if not visited[iAAB] or not visited[iABB]:
                visited[iAAB] = True
                visited[iABB] = True
                V[iAAB] = (2*pA + pB - np.dot(pB-pA, nA)*nA)/3
                V[iABB] = (2*pB + pA - np.dot(pA-pB, nB)*nB)/3
            if not visited[iBBC] or not visited[iBCC]:
                visited[iBBC] = True
                visited[iBCC] = True
                V[iBBC] = (2*pB + pC - np.dot(pC-pB, nB)*nB)/3
                V[iBCC] = (2*pC + pB - np.dot(pB-pC, nC)*nC)/3
            if not visited[iCCA] or not visited[iCAA]:
                visited[iCCA] = True
                visited[iCAA] = True
                V[iCCA] = (2*pC + pA - np.dot(pA-pC, nC)*nC)/3
                V[iCAA] = (2*pA + pC - np.dot(pC-pA, nA)*nA)/3
            e = (V[iAAB] + V[iABB] + V[iBBC] + V[iBCC] + V[iCCA] + V[iCAA])/6
            b = (pA+pB+pC)/3
            V[iABC] = e + (e-b)/2 
        return cls(V,F)

    def rasterize(self, res: int = 10, only_curves: bool = False):
        """
        Args:
            V (np.ndarray): Array of vertex positions (size Nx3)
            F (np.ndarray): Array of face indices (size Mx10)
            res (int, optional): Resolution of each P3 element. Defaults to 10.
            only_curves (bool, optional): if set to True, will only rasterize edges as Bezier curve and not draw the surface. Defaults to False.
            
        Returns:
            SurfaceMesh | Polyline: a mesh approximating the P3 mesh
        """
        if only_curves:
            return rasterize_pn_as_polylines(self.nodes, self.faces, [(0,3,4,1), (1,5,6,2), (2,7,8,0)], res)
        else:
            return self._rasterize_as_surface(res)

    def _rasterize_as_surface(self, res: int = 10):
        out = RawMeshData()

        def ij_to_p(i,j):
            return i*res - i*(i-1)//2 + j

        for triP3 in self.faces:
            pA,pB,pC,pAAB,pABB, pBBC, pBCC, pCCA,pCAA,pABC = (self.nodes[_x] for _x in triP3)
            paramU = np.linspace(0,1,res)
            paramV = np.linspace(0,1,res)
            f_ind = len(out.vertices)

            # evaluate vertices
            for i in range(res):
                for j in range(res-i):
                    out.vertices.append(evaluate_P3(pA,pB,pC,pAAB,pABB, pBBC, pBCC, pCCA, pCAA,pABC, paramU[i], paramV[j]))
                    
            # add faces
            for i in range(res-1):
                for j in range(res-1-i):
                    out.faces.append([f_ind+x for x in (ij_to_p(i,j), ij_to_p(i+1,j), ij_to_p(i,j+1))])
                    if j<res-2-i:
                        out.faces.append([f_ind+x for x in (ij_to_p(i+1,j), ij_to_p(i+1,j+1), ij_to_p(i,j+1))])
        return SurfaceMesh(out)


from ..mesh import from_arrays
from ..mesh.datatypes import PolyLine
from ..mesh.mesh_data import RawMeshData
from ..utils import iterators
import numpy as np

def chain_of_vertices(vertices: np.ndarray, loop:bool = False) -> PolyLine:
    """Creates a polyline that links the provided vertices in order.

    Args:
        vertices (np.ndarray): vertex positions.
        loop (bool, optional): whether to link the last vertex with the first, creating a closed loop. Defaults to False.

    Returns:
        PolyLine: _description_
    """
    pl = from_arrays(vertices, raw=True)
    n = len(pl.vertices)
    if loop:
        pl.edges += [x for x in iterators.cyclic_pairs(range(n))]
    else:
        pl.edges += [x for x in iterators.consecutive_pairs(range(n))]
    return PolyLine(pl)


def vector_field(origins: np.ndarray, vectors: np.ndarray, length_mult: float = 1.) -> PolyLine:
    """Creates the representation of a vector field from an array of origin points and an array of vectors.

    Args:
        origins (np.ndarray): size (N,K) with N the number of points and K<=3 the dimension. Origin of the vectors
        vectors (np.ndarray): size (N,K) with N the number of points and K<=3 the dimension. Coordinates of each vector.
        length_mult (float, optional): factor multiplied to each vector to modulate their length for vizualisation purposes. Defaults to 1.

    Raises:
        Exception: fails if the two arrays (origins and vectors) have a different shape
        Exception: fails if one of the two arrays have dimension > 3

    Returns:
        PolyLine: the vector field represented as a polyline.
    """
    pl = RawMeshData()
    
    # Sanitize input arrays
    origins, vectors = np.array(origins), np.array(vectors)
    if origins.shape != vectors.shape:
        raise Exception(f"Origin points array and vector array have different shapes, ({origins.shape} and {vectors.shape}).")
    if origins.shape[1]<3:
        origins = np.pad(origins, ((0,0),(0,3-origins.shape[1])))
        vectors = np.pad(vectors, ((0,0),(0,3-vectors.shape[1])))
    elif origins.shape[1]!=3: 
        raise Exception("Vertex array should have shape (n,3)")
    
    n = origins.shape[0]
    for i in range(n):
        end_i = origins[i] + length_mult*vectors[i]
        pl.vertices += [origins[i], end_i]
        pl.edges.append((2*i,2*i+1))
    return PolyLine(pl)


import numpy as np

from ..geometry import Vec
from ..mesh.datatypes import *
from ..mesh.mesh import _instanciate_raw_mesh_data
from ..mesh.mesh_data import RawMeshData

def triangle(P0: Vec, P1: Vec, P2: Vec) -> SurfaceMesh:
    """Generates a triangle from three vertices

    Args:
        P0,P1,P2 (Vec): three points

    Returns:
        SurfaceMesh: a triangle
    """
    out = RawMeshData()
    out.vertices += [P0, P1, P2]
    out.faces.append((0,1,2))
    return _instanciate_raw_mesh_data(out, 2)

def quad(P0: Vec, P1: Vec, P2: Vec, triangulate: bool = False) -> SurfaceMesh:
    """Generates a quad from three vertices
    
    ```
       P1-------
      /        /
     /        / 
    P0-------P2  
    ```

    Args:
        P0,P1,P2 (Vec): coordinates of corners. The fourth point is deduced as P2 + P1 - 2*P0
        triangulate (bool): whether to output two triangles instead of a quad. Defaults to False.
    
    Returns:
        SurfaceMesh: a quad
    """
    P0, P1, P2 = Vec(P0), Vec(P1), Vec(P2)
    P3 = P2 + P1 - P0

    out = RawMeshData()
    out.vertices += [P0, P1, P3, P2]
    if triangulate:
        out.faces += [(0,1,2), (0,2,3)]
    else:
        out.faces += [(0,1,2,3)]
    return _instanciate_raw_mesh_data(out,2)

def unit_grid(nu: int, nv: int, triangulate: bool = False, generate_uvs: bool=False) -> SurfaceMesh:
    """Generates a subdivided regular unit grid

    Args:
        nu (int): number of subdivisions on the horizontal axis
        nv (int): number of subdivisions on the vertical axis
        triangulate (bool, optional): whether to split quads into two triangles. Defaults to False.
        generate_uvs (bool, optional): whether to generate uv-coordinates. Defaults to False.

    Returns:
        SurfaceMesh: a subdivided unit grid of size nu*nv
    """
    out = RawMeshData()
    U = np.linspace(0,1,nu)
    V = np.linspace(0,1,nv)
    if generate_uvs:
        uv_attr = out.vertices.create_attribute("uv_coords",float,2)
    for i,u in enumerate(U):
        for j,v in enumerate(V):
            out.vertices.append(Vec(u,v,0))
            if generate_uvs:
                uv_attr[i*nu+j] = Vec(u,v)
            # generate faces
            if i<nu-1 and j<nv-1:
                if triangulate: # add two triangles
                    out.faces.append((i*nu+j, i*nu+j+1, (i+1)*nu+j))
                    out.faces.append((i*nu+j+1, (i+1)*nu+j+1, (i+1)*nu+j))
                else: # add a quad
                    out.faces.append((i*nu+j, i*nu+j+1, (i+1)*nu+j+1, (i+1)*nu+j))
    return _instanciate_raw_mesh_data(out, 2)


def unit_triangle(nu:int, nv:int, generate_uvs:bool=False) -> SurfaceMesh:
    """Generate a subdivided unit right triangle (half of a unit grid)

    Args:
        nu (int): number of subdivisions on the horizontal axis
        nv (int): number of subdivisions on the vertical axis
        generate_uvs (bool, optional): whether to generate uv-coordinates. Defaults to False.

    Returns
        SurfaceMesh: a subdivided unit triangle
    """
    out = RawMeshData()
    U = np.linspace(0,1,nu)
    V = np.linspace(1,0,nv)
    if generate_uvs:
        uv_attr = out.vertices.create_attribute("uv_coords",float,2)
    npt = 0
    for j,v in enumerate(V):
        for i,u in enumerate(U):
            if i>j: break
            out.vertices.append(Vec(u,v,0))
            if generate_uvs:
                uv_attr[npt] = Vec(u,v)
                npt += 1
    for j,v in enumerate(V):
        for i,u in enumerate(U):
            if i>j or j==nv-1: break
            kpt = j*(j+1)//2 + i
            if i<j: out.faces.append((kpt,kpt+j+2,kpt+1))
            out.faces.append((kpt,kpt+j+1,kpt+j+2))
    return _instanciate_raw_mesh_data(out, 2)


import numpy as np
import scipy.sparse as sp
import cmath

from ..geometry import *
from ..mesh.datatypes import *
from ..attributes.misc_faces import face_area
from ..attributes.misc_cells import cell_volume
from ..attributes.misc_corners import cotangent

##### For Surface, on vertices #####

@forbidden_mesh_types(PointCloud)
def graph_laplacian(mesh : Mesh) -> sp.csc_matrix:
    """Simplest Laplacian defined on a graph. uses uniform weights for connectivity.

    Parameters:
        mesh (Mesh): the input mesh

    Returns:
        (scipy.sparse.coo_matrix) : the laplacian operator as a sparse matrix
    """
    n = len(mesh.vertices)
    m = len(mesh.edges)
    ncoeffs = 2*m+n
    data, rows, cols = np.zeros(ncoeffs),np.zeros(ncoeffs, dtype=int),np.zeros(ncoeffs, dtype=int)
    
    def add(k, i, j, x):
        data[k] = x
        rows[k] = i
        cols[k] = j
        return k+1

    k = 0
    for l in mesh.id_vertices:
        adj = mesh.connectivity.vertex_to_vertices(l)
        k = add(k,l,l,len(adj))
        for b in adj:
            k = add(k, l, b, -1)
    return sp.csc_matrix((data, (rows,cols)), shape=(n,n))


@allowed_mesh_types(SurfaceMesh)
def area_weight_matrix(mesh : SurfaceMesh, inverse:bool = False, sqrt:bool = False, format:str="csc") -> sp.csc_matrix:
    """
    Returns the diagonal matrix A of area weights on vertices.
    
    Args:
        mesh (SurfaceMesh): input mesh
        inverse (bool, optional): whether to return A or A^-1. Defaults to False.
        sqrt (bool, optional): whether to return A or A^{1/2}. Can be combined with `inverse` to return A^{-1/2}. Defaults to False.
        format (str, optional): one of the sparse matrix format of scipy (csc, csr, coo, lil, ...). Defaults to csc.

    Returns:
        sp.csc_matrix: diagonal matrix of vertex areas
    """
    A = np.zeros(len(mesh.vertices))
    area = face_area(mesh)
    for iT,T in enumerate(mesh.faces):
        for u in T:
            A[u] += area[iT]
    if sqrt: A = np.sqrt(A)
    if inverse : A = 1/A # /!\ perform inverse after sqrt
    return sp.diags(A, format=format)

@allowed_mesh_types(SurfaceMesh)
def laplacian(
    mesh : SurfaceMesh, 
    cotan : bool=True,
    connection : "SurfaceConnectionVertices" = None, 
    order : int=4) -> sp.csc_matrix:
    """Cotan laplacian on vertices.

    Parameters:
        mesh (SurfaceMesh): input mesh
        cotan (bool) : whether to compute real cotan values for more precise discretization or only 0/1 values as a graph laplacian. Defaults to True.
        connection (SurfaceConnectionVertices, optional): For a laplacian on 1-forms, gives the angle in local bases of all adjacent edges. Defaults to None.
        order (int, optional): Order of the parallel transport (useful when computing frame fields). Does nothing if parallel_transport is set to None. Defaults to 4.

    Returns:
        scipy.sparse.csc_matrix : the Laplacian operator as a sparse matrix
    """
    n_coeffs = 12*len(mesh.faces)
    if cotan:
        if mesh.face_corners.has_attribute("cotan"):
            cot = mesh.face_corners.get_attribute("cotan")
        else:
            cot = cotangent(mesh)
    else:
        cot = None

    rows = np.zeros(n_coeffs, dtype=np.int32)
    cols = np.zeros(n_coeffs, dtype=np.int32)
    coeffs = np.zeros(n_coeffs, dtype=(complex if connection else np.float64))
    _c = 0
    
    for iT, (p,q,r) in enumerate(mesh.faces):
        if cotan:
            a,b,c = (cot[mesh.connectivity.vertex_to_corner_in_face(_v,iT)]/2 for _v in (p,q,r))
        else:
            a,b,c = 0.5, 0.5, 0.5
        for (i, j, v) in [(p, q, c), (q, r, a), (r, p, b)]:
            rows[_c], cols[_c], coeffs[_c], _c = i, i, v, _c+1
            rows[_c], cols[_c], coeffs[_c], _c = j, j, v, _c+1
            if connection is not None:
                ai, aj = connection.transport(i,j), connection.transport(j,i)
                rows[_c], cols[_c], coeffs[_c], _c = i, j, - v * cmath.rect(1., order*(ai - aj - math.pi)), _c+1
                rows[_c], cols[_c], coeffs[_c], _c = j, i, - v * cmath.rect(1., order*(aj - ai - math.pi)), _c+1
            else:
                rows[_c], cols[_c], coeffs[_c], _c = i, j, -v, _c+1
                rows[_c], cols[_c], coeffs[_c], _c = j, i, -v, _c+1 
    
    mat = sp.csc_matrix((coeffs,(rows,cols)), dtype= (complex if connection else np.float64))
    return mat

##### For Surface, on faces #####

@allowed_mesh_types(SurfaceMesh)
def cotan_edge_diagonal(mesh : SurfaceMesh, inverse:bool=True) -> sp.csc_matrix:
    """
    Builds a diagonal matrix of size |E|x|E| where the coefficients are the cotan weights, i.e.

        M[e,e] = 1/abs( cot(a_e) + cot(b_e))

    where a_e and b_e are the opposite angles in adjacent triangles of edge e.

    Args:
        mesh (SurfaceMesh): input mesh
        inverse (bool): whether to compute M or M^-1 (all coefficients on the diagonal inverted)

    Returns:
        sp.csc_matrix
    """
    m = len(mesh.edges)
    cotan = cotangent(mesh, persistent=False)
    coeffs = np.zeros(m)
    for ie,(u,v) in enumerate(mesh.edges):
        T1,uT1,vT1 = mesh.connectivity.direct_face(u,v,True)
        T2,vT2,uT2 = mesh.connectivity.direct_face(v,u,True)
        cT1,cT2 = 0., 0.
        if T1 is not None : 
            w = mesh.faces[T1][3-uT1-vT1]
            c = mesh.connectivity.vertex_to_corner_in_face(w,T1)
            cT1 = cotan[c]
        if T2 is not None : 
            w = mesh.faces[T2][3-uT2-vT2]
            c = mesh.connectivity.vertex_to_corner_in_face(w,T2)
            cT2 = cotan[c]
        
        if inverse:
            if abs(cT1 + cT2)<1e-8:
                coeffs[ie] = 1e8
            else:
                coeffs[ie] = 1/(cT1 + cT2)
        else:
            coeffs[ie] = cT1 + cT2
    return sp.diags(coeffs, format="csc")

@allowed_mesh_types(SurfaceMesh)
def area_weight_matrix_faces(mesh : SurfaceMesh, inverse : bool=False) -> sp.csc_matrix:
    """
    Returns the diagonal matrix A of area weights on faces
    Laplace-beltrami operator for a 2D manifold is (A^-1)L where A is the area weight and L is the cotan matrix

    Args:
        mesh (SurfaceMesh): the input mesh
        inverse (bool, optional): whether to return A or A^-1. Defaults to False.

    Returns:
        sp.csc_matrix
    """
    area = face_area(mesh, persistent=False).as_array()
    if inverse:
        area = 1/area
    return sp.diags(area, format="csc")

@allowed_mesh_types(SurfaceMesh)
def laplacian_triangles(
    mesh : SurfaceMesh, 
    cotan : bool = True, 
    connection : "SurfaceConnectionFaces" = None, 
    order : int=4) -> sp.lil_matrix:
    """
    Cotan laplacian defined on face connectivity (ie on the dual mesh)

    Args:
        mesh (SurfaceMesh): the supporting mesh
        cotan (bool, optional): whether to use cotan laplacian or . Defaults to True.
        connection (SurfaceConnectionFaces, optional): _description_. Defaults to None.
        order (int, optional): _description_. Defaults to 4.

    Returns:
        scipy.sparse.lil_matrix : the Laplacian operator as a sparse matrix
    """

    n = len(mesh.faces)
    m = len(mesh.edges)

    Nabla = sp.lil_matrix((m,n), dtype=complex if connection else np.float32) # gradient matrix
    if connection is not None:
        for ie,(ei,ej) in enumerate(mesh.edges):
            T1,T2 = mesh.connectivity.edge_to_faces(ei,ej)
            if T1 is not None and T2 is not None:
                Nabla[ie,T1] = -1
                Nabla[ie,T2] = cmath.rect(1, order*connection.transport(T1,T2))
    else:
         for ie,(ei,ej) in enumerate(mesh.edges):
            T1,T2 = mesh.connectivity.edge_to_faces(ei,ej)
            if T1 is not None and T2 is not None:
                Nabla[ie,T1] = -1
                Nabla[ie,T2] = 1
    Nabla = Nabla.tocsc()
    Nabla_star = Nabla.conj().transpose()
    if cotan:
        D = cotan_edge_diagonal(mesh)
        return Nabla_star @ D @ Nabla
    else:
        return Nabla_star @ Nabla

##### For Volumes #####

@allowed_mesh_types(VolumeMesh)
def volume_weight_matrix(mesh: VolumeMesh, inverse: bool = False, sqrt: bool = False, format: str = "csc") -> sp.dia_matrix:
    """
    Mass diagonal matrix for volume Laplacian.
    Args:
        mesh (VolumeMesh): input mesh
        inverse (bool, optional): whether to return A or A^-1. Defaults to False.
        sqrt (bool, optional): whether to return A or A^{1/2}. Can be combined with `inverse` to return A^{-1/2}. Defaults to False.
        format (str, optional): one of the sparse matrix format of scipy (csc, csr, coo, lil, ...). Defaults to csc.

    Returns:
        sp.csc_matrix: diagonal matrix of vertices area
    """
    A = np.zeros(len(mesh.vertices))
    volume = cell_volume(mesh, persistent=False)
    for iC,C in enumerate(mesh.cells):
        for u in C:
            A[u] += volume[iC] 
    if sqrt: A = np.sqrt(A)
    if inverse: A = 1/A
    return sp.diags(A, format=format)

@allowed_mesh_types(VolumeMesh)
def volume_laplacian(mesh : VolumeMesh) -> sp.lil_matrix:
    """Volume laplacian on vertices
    
    This is the 3D extension of the cotan laplacian, ie the discretization of the Laplace-Beltrami operator on 3D manifolds.

    Parameters:
        mesh (VolumeMesh): the input mesh

    Returns:
        scipy.sparse.lil_matrix: the Laplacian operator as a sparse matrix

    References:
       [1] https://cseweb.ucsd.edu/~alchern/projects/ConformalVolume/

       [2] https://www.cs.cmu.edu/~kmcrane/Projects/Other/nDCotanFormula.pdf
    """
    if not mesh.is_tetrahedral(): 
        raise NotImplementedError
    
    n = len(mesh.vertices)
    mat = sp.lil_matrix((n,n))
    for e,(I,J) in enumerate(mesh.edges):
        omega = 0
        for ic in mesh.connectivity.edge_to_cell(e):
            K,L = (x for x in mesh.cells[ic] if x not in (I,J)) # two other vertices of the tet
            l = distance(mesh.vertices[K], mesh.vertices[L])
            _,_,Z1 = face_basis(*(mesh.vertices[_u] for _u in (I,K,L)))
            _,_,Z2 = face_basis(*(mesh.vertices[_u] for _u in (J,L,K)))
            cot = abs(dot(Z1,Z2))/norm(cross(Z1,Z2))
            omega += l * cot / 6
        mat[I,I] += omega
        mat[J,J] += omega
        mat[I,J] = -omega
        mat[J,I] = -omega
    return mat

@allowed_mesh_types(VolumeMesh)
def laplacian_tetrahedra(mesh : VolumeMesh) -> sp.csc_matrix:
    """Laplacian defined on cell connectivity (ie on the dual volume mesh)

    Parameters:
        mesh (SurfaceMesh): input mesh

    Returns:
        scipy.sparse.csc_matrix
    """
    if not mesh.is_tetrahedral():
        raise Exception("Mesh is not tetrahedral")
    n = len(mesh.cells)
    mat = sp.lil_matrix((n,n))
    for c1 in mesh.id_cells:
        mat[c1,c1] += len(mesh.connectivity.cell_to_cell(c1))
        for c2 in mesh.connectivity.cell_to_cell(c1):
            mat[c1,c2] -= 1
    return mat.tocsc()


import numpy as np
from ..mesh.datatypes import *
from ..mesh.mesh_attributes import Attribute

@allowed_mesh_types(SurfaceMesh)
def generate_uv_colormap_vertices(
    mesh : SurfaceMesh, 
    vattr : Attribute, 
    vmin : float = None, 
    vmax : float = None
):
    """Generates the visualization of an attribute as a 1D UV-map onto the mesh, 
    so that a colormap can be used as a texture to visualize the values.
    Version for an attribute on vertices.

    Args:
        mesh (SurfaceMesh): considered mesh
        vattr (Attribute): the attribute on vertices
        vmin (float, optional): Minimum value for the map. If not provided, taken as the minimum value of the attribute. Defaults to None.
        vmax (float, optional): Maximum value for the map. If not provided, taken as the maximum value of the attribute. Defaults to None.
        res (int, optional): Resolution of the map. Defaults to 1000.
    """
    vmin = np.amin(vattr) if vmin is None else vmin
    vmax = np.amax(vattr) if vmax is None else vmax
    assert vmin <= vmax
    span = vmax - vmin
    uvs = mesh.face_corners.create_attribute("uv_coords", float, 2)

    for c,v in enumerate(mesh.face_corners):
        valv = max(min(vattr[v],vmax), vmin)
        valc = (valv - vmin) / span
        valc = max(min(valc, 0.999), 0.001)
        uvs[c] = [0, valc]
    return uvs

@allowed_mesh_types(SurfaceMesh)
def generate_uv_colormap_corners(
    mesh : SurfaceMesh, 
    cattr : Attribute, 
    vmin : float = None, 
    vmax : float = None
):
    """Generates the visualization of an attribute as a 1D UV-map onto the mesh, 
    so that a colormap can be used as a texture to visualize the values.
    Version for an attribute on face corners.

    Args:
        mesh (SurfaceMesh): considered mesh
        cattr (Attribute): the attribute on face corners
        vmin (float, optional): Minimum value for the map. If not provided, taken as the minimum value of the attribute. Defaults to None.
        vmax (float, optional): Maximum value for the map. If not provided, taken as the maximum value of the attribute. Defaults to None.
    """
    vmin = np.amin(cattr) if vmin is None else vmin
    vmax = np.amax(cattr) if vmax is None else vmax
    assert vmin <= vmax
    span = vmax - vmin
    uvs = mesh.face_corners.create_attribute("uv_coords", float, 2)

    for c in mesh.id_corners:
        valc = max(min(cattr[c],vmax), vmin)
        valc = (cattr[c] - vmin) / span
        uvs[c] = [0, valc]
    return uvs


@allowed_mesh_types(SurfaceMesh)
def generate_uv_colormap_faces(
    mesh : SurfaceMesh, 
    fattr : Attribute, 
    vmin : float = None,
    vmax : float = None
):
    """Generates the visualization of an attribute as a 1D UV-map onto the mesh, 
    so that a colormap can be used as a texture to visualize the values.
    Version for an attribute on faces.

    Args:
        mesh (SurfaceMesh): considered mesh
        cattr (Attribute): the attribute on face corners
        vmin (float, optional): Minimum value for the map. If not provided, taken as the minimum value of the attribute. Defaults to None.
        vmax (float, optional): Maximum value for the map. If not provided, taken as the maximum value of the attribute. Defaults to None.
    """
    vmin = np.amin(fattr) if vmin is None else vmin
    vmax = np.amax(fattr) if vmax is None else vmax
    assert vmin <= vmax
    span = vmax - vmin
    uvs = mesh.face_corners.create_attribute("uv_coords", float, 2)

    for c in mesh.id_corners:
        valf = max(min(fattr[mesh.connectivity.corner_to_face(c)],vmax), vmin)
        valc = (valf - vmin) / span
        uvs[c] = [0, valc]
    return uvs


NOT_AN_ID = 4294967295 # An integer (2^32 -1) representing a default value in geogram file format

complete_edges_from_faces = True
"""
In most file formats, surface meshes are given as a set of vertices and a set of faces, since edges can be retrieved from faces.
If this option is set to true, the edge set of the mesh will be constructed from the face information.
"""


complete_faces_from_cells = True
"""
In most file formats, volume meshes are given as a set of vertices and tetrahedra (cells).
If this option is set to true, the face set of the mesh will be constructed from the cells information.
"""

export_edges_in_obj = True
"""
Whether to export edges as `l <v> <v>` fields in .obj file format
"""

sort_neighborhoods = True
"""
If set to true, this will sort the corner connectivity arrays of a surface mesh, according to the computed half edges
"""

display_duplicate_attribute_warning = False
"""
When creating an attribute that already exists on a mesh, mouette prints a warning in the console and returns the attribute currently carrying this name.
If this flag is set to True, no warning is printed.
"""


import numpy as np
from math import pi, atan2

from ..mesh.datatypes import *
from ..mesh.mesh_attributes import Attribute, ArrayAttribute, Attribute
from .misc_faces import face_normals
from .misc_corners import corner_angles
from .interpolate import interpolate_faces_to_vertices
from .. import geometry as geom
from ..geometry import Vec

@forbidden_mesh_types(PointCloud)
def degree(mesh : Mesh, name : str = "degree", persistent:bool = True, dense:bool = True) -> Attribute :
    """Computes the degree of each vertex, that is the number of vertex that are adjacent.

    Parameters:
        mesh (Mesh): the input mesh. Pointclouds are forbidden
        name (str, optional): _description_. Defaults to "degree".
        persistent (bool, optional): If the attribute is persistent (stored in the mesh object) or not. Defaults to True.
        dense (bool, optional): Is the attribute dense (numpy array) or not (dict). Defaults to True

    Returns:
        Attribute(int) on vertices 
    """
    if persistent:
        deg = mesh.vertices.create_attribute(name, int, dense=dense)
    else:
        deg = ArrayAttribute(int, len(mesh.vertices)) if dense else Attribute(int)
    for (a,b) in mesh.edges:
        deg[a] = deg[a] + 1
        deg[b] = deg[b] + 1
    return deg

@allowed_mesh_types(SurfaceMesh)
def angle_defects(mesh : SurfaceMesh, zero_border=False, name = "angleDefect", persistent=True, dense=True) -> Attribute :
    """Computes the angle defect at each vertex, defined as 2*pi minus the sum of angles around the vertex.

    Angle defect is an discrete approximation of the gaussian curvature.
    This function relies on the computation of cotangents on the mesh.
    
    Warning: 
        Only works for triangulated meshes.

    Parameters:
        mesh (SurfaceMesh): the mesh
        zero_border (bool): if set to true, ignores the defect on the boundary of the mesh. Defaults to False.
        name (str, optional): Name given to the attribute. Defaults to "angleDefect".
        persistent (bool, optional): If the attribute is persistent (stored in the mesh object) or not. Defaults to True.
        dense (bool, optional): Is the attribute dense (numpy array) or not (dict). Defaults to True

    Returns:
        Attribute(float) on vertices

    Raises:
        Exception: fails if the mesh is not triangulated (ie if a face is not a triangle)
    """
    if not mesh.is_triangular():
        raise Exception("Tried to compute angle defects on a non-triangulated mesh")

    if persistent:
        defects = mesh.vertices.create_attribute(name, float, dense=dense, default_value=2*pi)
    else:
        defects = ArrayAttribute(float, len(mesh.vertices), default_value=2*pi) if dense else Attribute(float, default_value=2*pi)
    
    for i in mesh.boundary_vertices:
        defects[i] = 0 if zero_border else pi

    if mesh.face_corners.has_attribute("angles"):
        ang = mesh.face_corners.get_attribute("angles")
    else:
        ang = corner_angles(mesh, persistent=persistent)

    for C,V in enumerate(mesh.face_corners):
        if mesh.is_vertex_on_border(V) and zero_border: continue
        defects[V] -= ang[C]
    return defects

@allowed_mesh_types(SurfaceMesh)
def vertex_normals(mesh : SurfaceMesh, name="normals", persistent=True, interpolation="area", dense=True, custom_fnormals : Attribute = None) -> Attribute:
    """Computes normal directions as 3d vectors for each vertex.
    Normal at a vertex is a weighted sum of normals of adjacent faces. This function essentially interpolates the face normals.

    Parameters:
        mesh (SurfaceMesh): the input mesh
        name (str, optional): Name given to the attribute. Defaults to "normals".
        persistent (bool, optional): If the attribute is persistent (stored in the mesh object) or not. Defaults to True.
        interpolation (str, optional): Interpolation weighting mode.
            'uniform' : computes the mean for adjacent facet
            'area' : the mean is weighted by facet area. Default value.
            'angle' : the mean is weighted by internal angle at the vertex 
        dense (bool, optional): Is the attribute dense (numpy array) or not (dict). Defaults to True
        custom_fnormals (Attribute, optional): custom values for the face normals to be interpolated. Defaults to None.

    Raises:
        Exception: if 'interpolation' is not one of {'uniform', 'area', 'angles'}

    Returns:
        Attribute(float, 3) on vertices
    """

    if interpolation not in {'uniform', 'area', 'angle'}:
        raise Exception(f"vertex_normals: interpolation argument '{interpolation}' not recognized.\nPossibilites are {'uniform', 'area', 'angle'}.")

    if custom_fnormals is not None:
        fnormals = custom_fnormals
    elif mesh.faces.has_attribute("normals"):
        fnormals = mesh.faces.get_attribute("normals")
    else:
        fnormals = face_normals(mesh, persistent=persistent)
    
    if persistent:
        normals = mesh.vertices.create_attribute(name, float, 3, dense=dense)
    else:
        normals = ArrayAttribute(float, len(mesh.vertices), 3) if dense else Attribute(float)

    normals = interpolate_faces_to_vertices(mesh, fnormals, normals, weight=interpolation)
    for v in mesh.id_vertices:
        normals[v] = Vec.normalized(normals[v])
    return normals
    
@allowed_mesh_types(SurfaceMesh)
def border_normals(mesh : SurfaceMesh, name="borderNormals", persistent: bool = True, dense:bool = False) -> Attribute :
    """Computes the normal direction of the boundary curve.

    Parameters:
        mesh (SurfaceMesh): the mesh
        name (str, optional): Name given to the attribute. Defaults to "borderNormals".
        persistent (bool, optional): If the attribute is persistent (stored in the mesh object) or not. Defaults to True.
        dense (bool, optional): Is the attribute dense (numpy array) or not (dict). Defaults to True

    Returns:
        Attribute(float,3) on vertices
    """
    if persistent:
        bnormals = mesh.vertices.create_attribute(name, float, 3, dense=dense)
    else:
        bnormals = ArrayAttribute(float, len(mesh.vertices), 3) if dense else Attribute(float, 3)

    for v in mesh.boundary_vertices:
        v0 = mesh.connectivity.vertex_to_vertices(v)[0]
        v1 = mesh.connectivity.vertex_to_vertices(v)[-1]
        P,P0,P1 = (mesh.vertices[u] for u in (v,v0,v1))
        E0 = P - P0
        E1 = P1 - P
        T0 = mesh.connectivity.direct_face(v0,v)
        if T0 is None: T0 = mesh.connectivity.direct_face(v,v0)
        T1 = mesh.connectivity.direct_face(v,v1)
        if T1 is None: T1 = mesh.connectivity.direct_face(v1,v)
        _,_,N0 = geom.face_basis(*mesh.pt_of_face(T0))
        _,_,N1 = geom.face_basis(*mesh.pt_of_face(T1))
        bnormals[v] = geom.cross(E0,N0) + geom.cross(E1,N1)
        bnormals[v] = Vec.normalized(bnormals[v])
    return bnormals


"""
mouette.utils.iterators
Various iterators on list to make life easier
"""

def cyclic_pairs(L: list):
    """pairs of form (L[i], L[i+1]) that wraps at the end

    Parameters:
        L (list): list
    """
    n = len(L)
    for i in range(n):
        yield L[i],L[(i+1)%n]

def cyclic_pairs_enumerate(L : list):
    """pairs of form ((i,i+1),(L[i], L[i+1]) that wraps at the end

    Parameters:
        L (list): list
    """
    n = len(L)
    for i in range(n):
        yield (i, (i+1)%n), (L[i],L[(i+1)%n])


def cyclic_triplets(L : list):
    """
    triplets of form (L[i-1], L[i], L[i+1]) that wraps at the end

    Args:
        L (list): list
    """
    n = len(L)
    for i in range(n):
        yield L[i-1], L[i], L[(i+1)%n]

def consecutive_pairs(L : list):
    """pairs of form (L[i], L[i+1]) that does not wrap at the end

    Parameters:
        L (list): list
    """
    for i in range(len(L)-1):
        yield L[i],L[i+1]

def consecutive_triplets(L : list):
    """
    triplets of form (L[i-1], L[i], L[i+1]) that does not wraps at the end

    Args:
        L (list): list
    """
    n = len(L)
    for i in range(1, len(L)-1):
        yield L[i-1], L[i], L[i+1]

def cyclic_permutations(L):
    """All the cyclic permutations of elements of L

    Parameters:
        L (list): list
    """
    n = len(L)
    for j in range(n):
        yield [L[i - j] for i in range(n)]

def cyclic_perm_enumerate(L):
    """All the cyclic permutations of elements of L with the index of the first element.

    Parameters:
        L (list): list
    """
    n = len(L)
    for j in range(n):
        yield [((i-j)%n , L[i - j]) for i in range(n)]

def offset(L, k :int):
    n = len(L)
    return [L[(i+k)%n] for i in range(n)]


from ..mesh_data import RawMeshData
from ...geometry import Vec
from ... import config
from ...utils import keyify

def import_obj(path : str):
    """
    Parameters:
        path (str): the input file path
    """

    with open(path, 'r' ) as objf:
        output = parse_obj_data(objf.readlines())
    return output

def parse_vertex( vstr ):
    vals = vstr.split('/')
    vid = int(vals[0])-1
    tid = int(vals[1])-1 if len(vals)>1 and vals[1] else -1
    nid = int(vals[2])-1 if len(vals) > 2 else -1
    return (vid,tid,nid) 

def parse_obj_data(data):
    obj = RawMeshData()
    normals = []
    uv_coords = []
    faces = []
    for line in data:
        toks = line.split()
        if not toks: continue # empty line
        if toks[0] == 'v':
            obj.vertices.append( Vec([ float(v) for v in toks[1:4]]) ) # get only the first three components and ignore the rest
        elif toks[0] == 'vn':
            normals.append( Vec([ float(v) for v in toks[1:]]) )
        elif toks[0] == 'vt':
            uv_coords.append( Vec([float(toks[1]), float(toks[2])]) )
        elif toks[0] == 'f':
            faces.append([ parse_vertex(vstr) for vstr in toks[1:] ])
        elif toks[0] == 'l':
            v1,v2 = int(toks[1])-1, int(toks[2])-1
            e = keyify(v1,v2)
            obj.edges.append(e)

    normals_attr = obj.vertices.create_attribute("normals", float, 3)
    uv_attr = obj.face_corners.create_attribute("uv_coords", float, 2)
    ic = 0
    for iF,F in enumerate(faces):
        face = []
        for (vid,tid,nid) in F:
            face.append(vid)
            if nid!=-1:
                normals_attr[vid] = normals[nid]
            if tid!=-1:
                uv_attr[ic] = uv_coords[tid]
            ic += 1
        obj.faces.append(face)
        obj.face_corners += [(v,iF) for v in face]
    if normals_attr.empty() : obj.vertices.delete_attribute("normals")
    if uv_attr.empty(): obj.face_corners.delete_attribute("uv_coords")
    return obj

def export_obj(mesh, path):
    has_texcoords_vert = mesh.vertices.has_attribute("uv_coords") and not mesh.vertices.get_attribute("uv_coords").empty()
    has_texcoords_corners = mesh.face_corners.has_attribute("uv_coords") and not mesh.face_corners.get_attribute("uv_coords").empty()
    has_normals = mesh.vertices.has_attribute("normals") and not mesh.vertices.get_attribute("normals").empty()

    with open( path, 'w' ) as ofile:
        for vtx in mesh.vertices:
            ofile.write('v '+' '.join(['{}'.format(v) for v in vtx])+'\n')

        if has_texcoords_corners:
            texcoords = mesh.face_corners.get_attribute("uv_coords").as_array(len(mesh.face_corners))
            for tex in texcoords:
                ofile.write('vt '+' '.join([str(vt) for vt in tex])+'\n')
                    
        elif has_texcoords_vert:
            texcoords = mesh.vertices.get_attribute("uv_coords").as_array(len(mesh.vertices))
            for tex in texcoords:
                ofile.write('vt '+' '.join([str(vt) for vt in tex])+'\n')

        if has_normals :
            normals = mesh.vertices.get_attribute("normals")
            for i in mesh.id_vertices:
                nrm = normals[i]
                ofile.write('vn '+' '.join(['{}'.format(vn) for vn in nrm])+'\n')

        if config.export_edges_in_obj:
            if not config.complete_edges_from_faces:
                for a,b in mesh.edges:
                    ofile.write(f'l {a+1} {b+1}\n')
            elif mesh.edges.has_attribute("hard_edges"):
                for e in mesh.edges.get_attribute("hard_edges"):
                    a,b = mesh.edges[e]
                    ofile.write(f'l {a+1} {b+1}\n')

        cnr_id = 1
        for face in mesh.faces:
            str_face = ""
            for vid in face:
                str_id = str(vid+1)
                if has_texcoords_corners:
                    str_id+="/{}".format(cnr_id)
                elif has_texcoords_vert:
                    str_id+="/{}".format(vid+1)
                elif has_normals:
                    str_id+="/"
                if has_normals:
                    str_id+="/{}".format(vid+1)
                str_face += str_id + " "
                cnr_id += 1
            ofile.write('f ' + str_face + "\n")